# -*- coding: utf-8 -*-

### ------------------------------- Imports ------------------------------- ###
import os
import tqdm
import numpy as np
import pandas as pd
from seizyml.helper.event_match import get_szr_idx
### ----------------------------------------------------------------------- ###

def get_seizure_prop(parent_path, verified_predictions_dir, win):
    """
    Extracts seizure properties from verified prediction CSV files and saves the results to a new CSV.

    Parameters
    ----------
    parent_path : str
        Path to the parent directory containing verified predictions.
    verified_predictions_dir : str
        Subdirectory name with verified prediction CSV files.
    win : float
        Window size in seconds used to calculate seizure durations.

    Returns
    -------
    df : pd.DataFrame
        DataFrame containing seizure properties for each file.
    save_path : str
        File path to the saved CSV with seizure properties.
    """

    # get files
    ver_path = os.path.join(parent_path, verified_predictions_dir)
    filelist = list(filter(lambda k: '.csv' in k, os.listdir(ver_path)))
    
    # get columns
    cols = ['seizure_number','avg_seizure_dur_sec', 'total_time_seizing_sec',
            'coefficient_of_variation', 'recording_dur_hrs']
    
    # save array
    save_array = np.zeros([len(filelist), len(cols)])
    for i in tqdm.tqdm(range(len(filelist))): # loop through csv files
        
        # load predictions and get seizure bounds
        file_path = os.path.join(ver_path, filelist[i])
        ver_pred = np.loadtxt(file_path, delimiter=',')
        idx_bounds = get_szr_idx(ver_pred)
        
        # extract properties
        if idx_bounds.shape[0] > 0:
            save_array[i,0] = idx_bounds.shape[0]                              # seizure number
            save_array[i,1] = (np.sum(ver_pred)*win)/idx_bounds.shape[0]       # average seizure duration (seconds)     
            save_array[i,2] = (save_array[i,0]* save_array[i,1]).sum()         # total time seizing (seconds)
            iei = np.diff(idx_bounds[:,0])
            save_array[i,3] = 100*np.std(iei)/np.mean(iei)                     # inter-event-inteval
        save_array[i,4] = ver_pred.shape[0] * win/3600
    
    # create dataframe and save
    df = pd.DataFrame(data = save_array, columns = cols)    
    df.insert(0,'file_id', filelist) 
    df['seizures_per_hour'] = df['seizure_number'] / df['recording_dur_hrs'] 
    df['percent_time_seizing'] = 100*(df['total_time_seizing_sec']/3600) / df['recording_dur_hrs']
    save_path = os.path.join(parent_path, 'seizure_properties.csv')
    df.to_csv(save_path, index=False)

    return df, save_path






# -*- coding: utf-8 -*-
##### ----------------------------- IMPORTS ----------------------------- #####
import contextlib
import joblib
##### ------------------------------------------------------------------- #####

@contextlib.contextmanager
def tqdm_joblib(tqdm_object):
    """Context manager to patch joblib to report into tqdm progress bar given as argument"""

    def tqdm_print_progress(self):
        if self.n_completed_tasks > tqdm_object.n:
            n_completed = self.n_completed_tasks - tqdm_object.n
            tqdm_object.update(n=n_completed)

    original_print_progress = joblib.parallel.Parallel.print_progress
    joblib.parallel.Parallel.print_progress = tqdm_print_progress

    try:
        yield tqdm_object
    finally:
        joblib.parallel.Parallel.print_progress = original_print_progress
        tqdm_object.close()


# -*- coding: utf-8 -*-

### -------------- IMPORTS -------------- ###
import tables
### ------------------------------------- ###

def load_data(load_path):
    """
    Retrieve h5 data.

    Parameters
    ----------
    main_path : Str

    Returns
    -------
    data : ndarray (1d = segments, 2d = time, 3d = channels)

    """
    
    # load lfp/eeg data
    f = tables.open_file(load_path, mode='r') # open tables object
    data = f.root.data[:]; f.close() # load data
    del f
    return data

def save_data(save_path, data):
    """
    Save h5 data

    Parameters
    ----------
    save_path : Str
    data : ndarray

    Returns
    -------

    """
    
    try:
        # Saving Parameters
        atom = tables.Float64Atom() # declare data type 
        fsave = tables.open_file(save_path , mode='w') # open tables object
        
        # create data store 
        if len(data.shape) == 3:
            ds = fsave.create_earray(fsave.root, 'data', atom,
                                        [0, data.shape[1], data.shape[2]])
        elif len(data.shape) == 2:
            ds = fsave.create_earray(fsave.root, 'data', atom,
                                        [0, data.shape[1]])    
        elif len(data.shape) == 1:
            ds = fsave.create_earray(fsave.root, 'data', atom,
                                        [0])
                
        ds.append(data) # append data
        fsave.close() # close tables object
        del fsave
        return 1
    
    except Exception as e:
        print('File could not be saved')
        print(e)
        return 0


# settings_editor.py
import customtkinter as ctk
import tkinter as tk
from tkinter import messagebox

class SettingsEditor(ctk.CTk):
    def __init__(self, current_settings):
        super().__init__()
        # This attribute will hold updated settings (or None if the user cancels)
        self.new_settings = None
        # Work on a copy so we don't modify the original until saving
        self.settings = current_settings.copy()

        self.title("Settings Editor")
        self.geometry("600x600")
        
        # Create a Tabview widget for grouping settings
        self.tabview = ctk.CTkTabview(self, width=580, height=500)
        self.tabview.pack(pady=10, padx=10)
        
        # Add tabs
        self.tabview.add("Data Settings")
        self.tabview.add("GUI Settings")
        self.tabview.add("Features")
        self.tabview.add("Post Processing")
        self.tabview.add("Directories")
        
        # Build content for each tab
        self.create_data_settings_tab()
        self.create_gui_settings_tab()
        self.create_features_tab()
        self.create_post_processing_tab()
        self.create_directories_tab()
        
        # Frame for Save and Cancel buttons
        self.button_frame = ctk.CTkFrame(self, width=580, height=50)
        self.button_frame.pack(pady=10)
        self.save_button = ctk.CTkButton(self.button_frame, text="Save", command=self.save_settings)
        self.save_button.pack(side="left", padx=10)
        self.cancel_button = ctk.CTkButton(self.button_frame, text="Cancel", command=self.cancel)
        self.cancel_button.pack(side="left", padx=10)

    def create_data_settings_tab(self):
        tab = self.tabview.tab("Data Settings")
        # Channels
        ctk.CTkLabel(tab, text="Channels (comma separated):").grid(row=0, column=0, padx=10, pady=5, sticky="w")
        self.channels_entry = ctk.CTkEntry(tab, width=300)
        self.channels_entry.grid(row=0, column=1, padx=10, pady=5)
        self.channels_entry.insert(0, ', '.join(self.settings['channels']))
        
        # Sampling frequency (fs)
        ctk.CTkLabel(tab, text="Sampling frequency (fs):").grid(row=1, column=0, padx=10, pady=5, sticky="w")
        self.fs_entry = ctk.CTkEntry(tab, width=100)
        self.fs_entry.grid(row=1, column=1, padx=10, pady=5, sticky="w")
        self.fs_entry.insert(0, str(self.settings['fs']))
        
        # Window length (win)
        ctk.CTkLabel(tab, text="Window length (win):").grid(row=2, column=0, padx=10, pady=5, sticky="w")
        self.win_entry = ctk.CTkEntry(tab, width=100)
        self.win_entry.grid(row=2, column=1, padx=10, pady=5, sticky="w")
        self.win_entry.insert(0, str(self.settings['win']))
    
    def create_gui_settings_tab(self):
        tab = self.tabview.tab("GUI Settings")
        ctk.CTkLabel(tab, text="GUI Window (gui_win):").grid(row=0, column=0, padx=10, pady=5, sticky="w")
        self.gui_win_entry = ctk.CTkEntry(tab, width=100)
        self.gui_win_entry.grid(row=0, column=1, padx=10, pady=5, sticky="w")
        self.gui_win_entry.insert(0, str(self.settings['gui_win']))
    
    def create_features_tab(self):
        tab = self.tabview.tab("Features")
        ctk.CTkLabel(tab, text="Features (comma separated):").grid(row=0, column=0, padx=10, pady=5, sticky="w")
        self.features_entry = ctk.CTkEntry(tab, width=400)
        self.features_entry.grid(row=0, column=1, padx=10, pady=5)
        self.features_entry.insert(0, ', '.join(self.settings['features']))
        
        ctk.CTkLabel(tab, text="Feature Select Threshold:").grid(row=1, column=0, padx=10, pady=5, sticky="w")
        self.feature_thresh_entry = ctk.CTkEntry(tab, width=100)
        self.feature_thresh_entry.grid(row=1, column=1, padx=10, pady=5, sticky="w")
        self.feature_thresh_entry.insert(0, str(self.settings['feature_select_thresh']))
        
        ctk.CTkLabel(tab, text="Feature Size (comma separated):").grid(row=2, column=0, padx=10, pady=5, sticky="w")
        self.feature_size_entry = ctk.CTkEntry(tab, width=200)
        self.feature_size_entry.grid(row=2, column=1, padx=10, pady=5, sticky="w")
        self.feature_size_entry.insert(0, ', '.join(map(str, self.settings['feature_size'])))
        
        ctk.CTkLabel(tab, text="N least correlated:").grid(row=3, column=0, padx=10, pady=5, sticky="w")
        self.nleast_corr_entry = ctk.CTkEntry(tab, width=100)
        self.nleast_corr_entry.grid(row=3, column=1, padx=10, pady=5, sticky="w")
        self.nleast_corr_entry.insert(0, str(self.settings['nleast_corr']))
    
    def create_post_processing_tab(self):
        tab = self.tabview.tab("Post Processing")
        ctk.CTkLabel(tab, text="Post Processing Method:").grid(row=0, column=0, padx=10, pady=5, sticky="w")
        self.pp_method_combobox = ctk.CTkComboBox(tab, values=["dual_threshold", "other_method"])
        self.pp_method_combobox.grid(row=0, column=1, padx=10, pady=5, sticky="w")
        self.pp_method_combobox.set(self.settings['post_processing_method'])
        
        ctk.CTkLabel(tab, text="Rolling Window:").grid(row=1, column=0, padx=10, pady=5, sticky="w")
        self.rolling_window_entry = ctk.CTkEntry(tab, width=100)
        self.rolling_window_entry.grid(row=1, column=1, padx=10, pady=5, sticky="w")
        self.rolling_window_entry.insert(0, str(self.settings['rolling_window']))
        
        ctk.CTkLabel(tab, text="Event Threshold:").grid(row=2, column=0, padx=10, pady=5, sticky="w")
        self.event_threshold_entry = ctk.CTkEntry(tab, width=100)
        self.event_threshold_entry.grid(row=2, column=1, padx=10, pady=5, sticky="w")
        self.event_threshold_entry.insert(0, str(self.settings['event_threshold']))
        
        ctk.CTkLabel(tab, text="Boundary Threshold:").grid(row=3, column=0, padx=10, pady=5, sticky="w")
        self.boundary_threshold_entry = ctk.CTkEntry(tab, width=100)
        self.boundary_threshold_entry.grid(row=3, column=1, padx=10, pady=5, sticky="w")
        self.boundary_threshold_entry.insert(0, str(self.settings['boundary_threshold']))
        
        ctk.CTkLabel(tab, text="Dilation:").grid(row=4, column=0, padx=10, pady=5, sticky="w")
        self.dilation_entry = ctk.CTkEntry(tab, width=100)
        self.dilation_entry.grid(row=4, column=1, padx=10, pady=5, sticky="w")
        self.dilation_entry.insert(0, str(self.settings['dilation']))
        
        ctk.CTkLabel(tab, text="Erosion:").grid(row=5, column=0, padx=10, pady=5, sticky="w")
        self.erosion_entry = ctk.CTkEntry(tab, width=100)
        self.erosion_entry.grid(row=5, column=1, padx=10, pady=5, sticky="w")
        self.erosion_entry.insert(0, str(self.settings['erosion']))
    
    def create_directories_tab(self):
        tab = self.tabview.tab("Directories")
        ctk.CTkLabel(tab, text="Data Directory:").grid(row=0, column=0, padx=10, pady=5, sticky="w")
        self.data_dir_entry = ctk.CTkEntry(tab, width=200)
        self.data_dir_entry.grid(row=0, column=1, padx=10, pady=5, sticky="w")
        self.data_dir_entry.insert(0, self.settings['data_dir'])
        
        ctk.CTkLabel(tab, text="Processed Directory:").grid(row=1, column=0, padx=10, pady=5, sticky="w")
        self.processed_dir_entry = ctk.CTkEntry(tab, width=200)
        self.processed_dir_entry.grid(row=1, column=1, padx=10, pady=5, sticky="w")
        self.processed_dir_entry.insert(0, self.settings['processed_dir'])
        
        ctk.CTkLabel(tab, text="Model Predictions Directory:").grid(row=2, column=0, padx=10, pady=5, sticky="w")
        self.model_predictions_entry = ctk.CTkEntry(tab, width=200)
        self.model_predictions_entry.grid(row=2, column=1, padx=10, pady=5, sticky="w")
        self.model_predictions_entry.insert(0, self.settings['model_predictions_dir'])
        
        ctk.CTkLabel(tab, text="Verified Predictions Directory:").grid(row=3, column=0, padx=10, pady=5, sticky="w")
        self.verified_predictions_entry = ctk.CTkEntry(tab, width=200)
        self.verified_predictions_entry.grid(row=3, column=1, padx=10, pady=5, sticky="w")
        self.verified_predictions_entry.insert(0, self.settings['verified_predictions_dir'])
        
        ctk.CTkLabel(tab, text="Trained Model Directory:").grid(row=4, column=0, padx=10, pady=5, sticky="w")
        self.trained_model_entry = ctk.CTkEntry(tab, width=200)
        self.trained_model_entry.grid(row=4, column=1, padx=10, pady=5, sticky="w")
        self.trained_model_entry.insert(0, self.settings['trained_model_dir'])
    
    def save_settings(self):
        try:
            new_settings = {}
            # Data Settings
            new_settings['channels'] = [s.strip() for s in self.channels_entry.get().split(',') if s.strip()]
            new_settings['fs'] = int(self.fs_entry.get())
            new_settings['win'] = int(self.win_entry.get())
            
            # GUI Settings
            new_settings['gui_win'] = int(self.gui_win_entry.get())
            
            # Features
            new_settings['features'] = [s.strip() for s in self.features_entry.get().split(',') if s.strip()]
            new_settings['feature_select_thresh'] = float(self.feature_thresh_entry.get())
            new_settings['feature_size'] = [int(s.strip()) for s in self.feature_size_entry.get().split(',') if s.strip()]
            new_settings['nleast_corr'] = int(self.nleast_corr_entry.get())
            
            # Post Processing
            new_settings['post_processing_method'] = self.pp_method_combobox.get()
            new_settings['rolling_window'] = int(self.rolling_window_entry.get())
            new_settings['event_threshold'] = float(self.event_threshold_entry.get())
            new_settings['boundary_threshold'] = float(self.boundary_threshold_entry.get())
            new_settings['dilation'] = int(self.dilation_entry.get())
            new_settings['erosion'] = int(self.erosion_entry.get())
            
            # Directories
            new_settings['data_dir'] = self.data_dir_entry.get()
            new_settings['processed_dir'] = self.processed_dir_entry.get()
            new_settings['model_predictions_dir'] = self.model_predictions_entry.get()
            new_settings['verified_predictions_dir'] = self.verified_predictions_entry.get()
            new_settings['trained_model_dir'] = self.trained_model_entry.get()
            
            self.new_settings = new_settings  # Save the new settings
            self.destroy()  # Close the GUI
        except Exception as e:
            messagebox.showerror("Error", f"An error occurred: {e}")

    def cancel(self):
        # If the user cancels, simply close the GUI without saving changes.
        self.destroy()


def edit_settings_gui(current_settings):
    """
    Launch the settings editor GUI with the current settings.
    Returns the updated settings if saved, or the original settings if canceled.
    """
    app = SettingsEditor(current_settings)
    app.mainloop()
    # Return the new settings if they exist, else the original settings.
    return app.new_settings if app.new_settings is not None else current_settings

# ----------------------------------------------------------------



# -*- coding: utf-8 -*-

### ------------------------ IMPORTS ---------------------------- ###
import tables
import numpy as np
from scipy import signal
import matplotlib.pyplot as plt
### ------------------------------------------------------------- ###

def generate_noisy_sine_waves(num_samples, num_channels, sampling_rate, frequency=10, noise_level=1):
    """
    Generate noisy sine waves.
    
    Parameters
    ----------
    num_samples : int
        Number of samples in the generated sine waves.
    num_channels : int
        Number of channels to generate (equivalent to the number of columns in the output).
    sampling_rate : int
        The sampling rate in Hz.
    frequency : int or float, optional
        The frequency of the sine waves in Hz (default is 10 Hz).
    noise_level : float, optional
        The amplitude of random noise to be added to the sine waves (default is 1).
    
    Returns
    -------
    time : numpy.ndarray
        The time vector corresponding to the generated sine waves.
    noisy_sine_waves : numpy.ndarray
        The generated noisy sine waves of shape (num_samples, num_channels).
        Each column represents a sine wave with added noise.
    
    Notes
    -----
    This function generates noisy sine waves with the specified parameters. It creates `num_channels` sine waves
    with a frequency of `frequency` Hz. The sine waves are sampled at a rate of `sampling_rate` Hz to create
    `num_samples` data points.
    
    Random noise is added to each sine wave with an amplitude defined by `noise_level`. The noise makes the
    generated signals look more biological and realistic.
    
    Example
    -------
    >>> time, data = generate_noisy_sine_waves(num_samples=1000, num_channels=5, sampling_rate=1000, frequency=5, noise_level=0.2)
    """
    time = np.arange(num_samples) / sampling_rate

    # Create the sine waves for each channel
    sine_waves = np.zeros((num_samples, num_channels))
    for channel in range(num_channels):
        sine_waves[:, channel] = np.sin(2 * np.pi * frequency * time)

    # Add some random noise to the sine waves
    noise = np.random.normal(0, noise_level, size=(num_samples, num_channels))
    noisy_sine_waves = sine_waves + noise

    return time, noisy_sine_waves


# Here we are creating a datastore in append mode, so that we can load large files in chuncks, 
# downsample them and then store them in h5 file format. When we finish looping through the file we close the datastore.
# After that we can access our h5 file. 

# =====================
# 1. Basic Settings
# =====================

# Here, we setup the basic configuration parameters for our data processing
file_name = 'converted_data.h5'
fs = 1000  # original sampling rate
new_fs = 100  # new sampling rate after decimation
win = 5  # window size in seconds for data storage
downsample_factor = int(fs/new_fs)  # compute the downsample factor

# =======================
# 2. Channel Configurations
# =======================

# We define how many channels our data will have, and which channels we're interested in
total_channels = 5  # total number of channels in the generated data
select_channels = [1, 4]  # list of channels to be selected for decimation

# ==========================
# 3. Datastore Shape Settings
# ==========================

# Here we set up the shape of our datastore.
chunksize = 10000  # number of rows with a bin size of (win*new_fs). **This would define how much data will be loaded.***
cols = int(win*new_fs)  # number of columns in each data row
chunkshape = [chunksize, cols, len(select_channels)]  # shape of each chunk for data storage
datastore_shape = [0, cols, len(select_channels)]  # shape of the complete data

# ==========================
# 4. Create HDF5 Datastore
# ==========================

# Next, we create a HDF5 datastore to save our data.
fsave = tables.open_file(file_name, mode='w')  # open an HDF5 file in write mode
ds = fsave.create_earray(fsave.root, 'data', tables.Float64Atom(),
                         datastore_shape, chunkshape=chunkshape)  # create an extendable array (earray) for data storage

# ==========================
# 5. Generate and Load Data
# ==========================

# Here we generate and load data. In a real use case, this would be replaced by actual data loading in a loop
_, data_chunk = generate_noisy_sine_waves(chunksize*fs, total_channels, fs, frequency=10, noise_level=.5)

# ==========================
# 6. Decimate the Data
# ==========================

# We decimate the data from the selected channels
decimate_data = []
for i in select_channels:
    data_ch = signal.decimate(data_chunk[:, i], downsample_factor)
    decimate_data.append(data_ch)
decimate_data = np.vstack(decimate_data).T

# ==========================
# 7. Reshape and Store the Data
# ==========================

# Finally, we reshape the decimated data and append it to our HDF5 datastore
data = decimate_data.reshape(-1, cols, len(select_channels))  # reshape the data to fit the storage shape
ds.append(data)  # append the reshaped data to the HDF5 datastore

# ==========================
# 8. Close the Datastore
# ==========================
fsave.close()


# ------------------------------------------------------------------------- #

# ====================================
# 1. Load Data from the HDF5 Datastore
# ====================================

# Here we open the HDF5 file in read mode, and then load all data into memory
f = tables.open_file(file_name, mode = 'r') # open tables object
load_data = f.root.data[:] # load data
f.close() # always remember to close the file after you're done with it

# =======================================
# 2. Prepare Data for Visualization
# =======================================

# We select a portion of the data that we're interested in visualizing
# Here we select data from the first channel (0) and the first window size 
y = data_chunk[0:(win*fs), select_channels[0]]
y2 = load_data[0,:,0]

# Prepare the time axis for both original and downsampled data
t = np.arange(0,len(y),1) # time axis for the original data
t2 = np.arange(0,len(y),10) # time axis for the downsampled data

# =======================================
# 3. Data Visualization
# =======================================

# Finally, we plot our original data (grey line) and the data we loaded from the HDF5 datastore (black line)
plt.plot(t, y, label='saved_data', linewidth=1, color='grey')
plt.plot(t2, y2, label='loaded_data', linewidth=2, color='black')
plt.legend()







# -*- coding: utf-8 -*-

### ---------------------- IMPORTS ---------------------- ###
import os
import numpy as np
import matplotlib.pyplot as plt
from matplotlib.widgets import SpanSelector
from PyQt5 import QtCore
### ----------------------------------------------------- ###

class VerifyGui(object):
    """
        Matplotlib GUI for user seizure verification.
    """
    
    def __init__(self, parent_path, settings, file_id, data, idx_bounds, color_array=None):
        """  

        Parameters
        ----------
        parent_path, str, path parent directory
        settings : dict, with configuration settings
        file_id : str, file name
        data : 3D Numpy array, (1D = seizure segments, 2D =  columns (samples: window*sampling rate), 3D = channels) 
        idx_bounds : 2D Numpy array (1D = seizure segments, 2D, 1 = start, 2 = stop index)
        color_array: array-like, with colors to plot depending on verification status (accepted, rejected, unverified)

        Returns
        -------
        None.

        """
        
        # settings
        self.file_id = file_id                                                  # file name
        self.model_win = settings['win']                                        # model window in seconds
        self.gui_win = settings['gui_win']                                      # gui window in seconds
        self.fs = settings['fs']                                                # sampling rate
        self.bounds = 60                                                        # surrounding region in seconds
        self.bounds_bins = round(self.bounds/self.gui_win)                      # get surround time in bins
        self.ch_list = np.array(settings['channels'])                           # channel names
        self.verpred_dir = os.path.join(parent_path, settings['verified_predictions_dir'])
        self.accepted_color = 'palegreen'
        self.rejected_color = 'salmon'
        self.unverified_color = 'w'
        self.wait_time = 0.1                                                    # in seconds
        self.ind = 0

        # check if data were verified
        if color_array is None:
            self.color_array = [self.unverified_color]*idx_bounds.shape[0]
        else:
            self.color_array = list(color_array)
        
        # adjust index to match gui window
        self.data = data.reshape(-1, int(self.fs * self.gui_win), data.shape[2])
        self.idx = (np.copy(idx_bounds) * (self.model_win / self.gui_win)).astype(int)

        # create figure and axis
        self.fig, self.axs = plt.subplots(data.shape[2], 1, sharex=True, figsize=(12,8))
        if data.shape[2] == 1:
            self.axs = [self.axs]
        for i in range(self.axs.shape[0]): 
            self.axs[i].spines["top"].set_visible(False)
            self.axs[i].spines["right"].set_visible(False)
            self.axs[i].spines["bottom"].set_visible(False)
        self.plot_data()
           
        # connect callbacks and add key legend 
        plt.subplots_adjust(bottom=0.15)
        self.fig.suptitle('To Select boundaries drag mouse : '+ self.file_id, fontsize=12)         # title  
        self.fig.text(0.5, 0.09,'Time  (Seconds)', ha="center")         # xlabel
        self.fig.text(.02, .5, 'Amp. (V)', ha='center', va='center', rotation='vertical')          # ylabel
        self.fig.text(0.5, 0.04, 
                      "** Accept/Reject = a/r,      Previous/Next = ←/→,    ctrl+Previous/Next = -10/+10, \n Enter = Save, Esc = close(no Save), Drag cursor (left click) to adjust bounds**" ,
                      ha="center", bbox=dict(boxstyle="square", ec=(1., 1., 1.), fc=(0.9, 0.9, 0.9),))
        self.fig.canvas.callbacks.connect('key_press_event', self.keypress)
        
        # disable x button
        win = plt.gcf().canvas.manager.window
        win.setWindowFlags(win.windowFlags() | QtCore.Qt.CustomizeWindowHint)
        win.setWindowFlags(win.windowFlags() & ~QtCore.Qt.WindowCloseButtonHint)
        
        # span selector
        _ = SpanSelector(self.axs[0], self.onselect, 'horizontal', useblit=True,
            rectprops=dict(alpha=0.5, facecolor='tab:blue'))
        plt.show()
    
    @staticmethod
    def get_hours(seconds):
        hours = seconds // 3600
        minutes = (seconds % 3600) // 60
        seconds = seconds % 60
        time_str = '{}:{}:{}'.format(int(hours), int(minutes), int(seconds))
        return time_str
          
    def save_idx(self):
        """
        Save user predictions to csv file as binary
        Returns
        -------
        None.
        """
        # create bool array with verified predictions
        ver_pred = np.zeros(self.data.shape[0])
        for idx, color in zip(self.idx, self.color_array):
            if color == self.accepted_color:
                ver_pred[idx[0]:idx[1]+1] = 1
            
        # save file
        np.savetxt(os.path.join(self.verpred_dir, self.file_id), ver_pred, delimiter=',',fmt='%i')
        np.savetxt(os.path.join(self.verpred_dir, 'color_'+self.file_id.replace('.csv', '.txt')), np.array(self.color_array), delimiter=',', fmt='%s')
        print('Verified predictions for ', self.file_id, ' were saved.\n')
    
    ## ----- User Selection ----##        
    def onselect(self, xmin, xmax):
        """
        Parameters
        ----------
        xmin : Float
            Xmin-user selection.
        xmax : Float
            Xmax-user selection.
        """
               
        # find user segment index from plot
        indmin = int(xmin)
        indmax = int(xmax)
        
        # pass to index
        self.idx[self.i,0] = indmin
        self.idx[self.i,1] = indmax
        
        # highlight user selected region
        self.plot_data(user_start=indmin, user_stop=indmax)
            
    def plot_data(self, user_start=None, user_stop=None):
        """
        Plot channels with highlighted seizures.

        Returns
        -------
        None.

        """

        # get index, start and stop times
        self.i = self.ind % self.idx.shape[0]
        start,stop = self.idx[self.i]
        if user_start is not None:
            start,stop = user_start, user_stop
        
        # get seizure time
        timestr = VerifyGui.get_hours(start*self.gui_win)
        timestr = '#' + str(self.i+1) + ' - '+ timestr

        # plot channels
        for i in range(self.axs.shape[0]): 
            start_idx = max(0, start - self.bounds_bins)
            stop_idx = min(self.data.shape[0], stop + self.bounds_bins)
            y = self.data[start_idx:stop_idx, :, i].flatten()
            t = np.linspace(start - self.bounds_bins, stop + self.bounds_bins, len(y))*self.gui_win
            self.axs[i].clear()
            self.axs[i].plot(t, y, color='k', linewidth=0.75, alpha=0.9, label=timestr)
            self.axs[i].set_facecolor(self.color_array[self.i])
            self.axs[i].legend(loc='upper right')
            self.axs[i].set_title(self.ch_list[i], loc='left')
            
            # plot highlighted region
            yzoom = self.data[start: stop+1,:,i].flatten()
            tzoom = np.linspace(start, stop+1, len(yzoom))*self.gui_win
            self.axs[i].plot(tzoom, yzoom, color='orange', linewidth=0.75, alpha=0.9)
        self.fig.canvas.draw()
       
    ## ------  Keyboard press ------ ##     
    def keypress(self, event):

        # navigate one event
        if event.key == 'right':
            self.ind += 1
            self.plot_data()
        
        if event.key == 'left':
            self.ind -= 1
            self.plot_data()
            
        # navigate ten events
        if event.key == 'ctrl+right':
            self.ind += 10
            self.plot_data()
        
        if event.key == 'ctrl+left':
            self.ind -= 10
            self.plot_data()
        
        # accept and reject events
        if event.key == 'a':
            self.color_array[self.i] = self.accepted_color
            self.plot_data()
            self.fig.canvas.draw()
            plt.pause(self.wait_time)
            self.ind += 1
            self.plot_data()
                
        if event.key == 'r':
            self.color_array[self.i] = self.rejected_color
            self.plot_data()
            self.fig.canvas.draw()
            plt.pause(self.wait_time)
            self.ind += 1
            self.plot_data()
        
        if event.key == 'enter': 
            plt.close()
            self.save_idx()
            
            # get number of accepted, rejected, and unverified seizures
            verified_array = np.array(self.color_array)
            not_verified = np.sum(verified_array == self.unverified_color)
            if not_verified > 0:
                print(f'--> Warning {not_verified} were not verified!')
            total_seizures = len(verified_array)
            accepted = np.sum(verified_array== self.accepted_color)
            rejected = np.sum(verified_array == self.rejected_color)
            print(f'Total predicted events = {total_seizures}. Accepted = {accepted}, Rejected = {rejected}, Unverified = {not_verified}.\n')
            
        if event.key == 'escape': 
            plt.close()



# -*- coding: utf-8 -*-

### ------------------- Imports ------------------- ###
from seizyml.helper import features
import numpy as np
from joblib import Parallel, delayed
import multiprocessing
n_jobs = int(multiprocessing.cpu_count() * 0.8)
### ----------------------------------------------- ###

def compute_features(data, single_channel_functions, channel_names, fs):
    """
    Compute features from given 3D data.
    
    Parameters
    ----------
    data : ndarray, 3D array of shape (segments, time, channels) containing the input data.
    single_channel_functions : List of function names for single channel features, defined in features.py.
    channel_names : List of channel names used to name the features.
    fs: int, sampling rate

    Returns
    -------
    features_array : 2D array containing the computed features for each segment, with shape (segments, num_features).
    feature_labels : Array of labels for the computed features, corresponding to the columns in features_array.
    """
    
    # get dims and init lists
    num_segments, _, num_channels = data.shape
    feature_labels = []
    features_list = []

    # Calculate single channel features
    param_list = [getattr(features, func_name) for func_name in single_channel_functions]
    def extract_features_for_segment(segment, param_list):
        return [func(segment, fs=fs) for func in param_list]
        
    for c in range(num_channels):
        x_data = Parallel(n_jobs=n_jobs)(delayed(extract_features_for_segment)(segment, param_list) for segment in data[:,:,c])
        features_list.append(np.array(x_data))
        feature_labels.extend([f"{func_name}-{channel_names[c]}" for func_name in single_channel_functions])

    features_array = np.column_stack(features_list)
    return features_array, np.array(feature_labels)

def compute_selected_features(data, selected_feature_names, channel_names, fs):
    """
    Compute selected features from given 3D data based on feature names.

    Parameters
    ----------
    data : 3D array of shape (segments, time, channels) containing the input data.
    selected_feature_names : List of selected feature names, must match the naming convention used in compute_features.
    channel_names : List of channel names used to match the features.
    fs: int, sampling rate
    
    Returns
    -------
    features_array : 2D array containing the computed features for each segment, with shape (segments, num_selected_features).
    feature_labels : Array of labels for the computed features, corresponding to the columns in features_array.
    """
    
    # get dims and init lists
    num_segments, _, num_channels = data.shape
    feature_labels = []
    features_list = []

    for feature_name in selected_feature_names:
        
        # separate feature name to function and channel
        parts = feature_name.split('-')
        func_name = parts[0]
        channel_indices = [channel_names.index(ch_name) for ch_name in parts[1:]]
        func = getattr(features, func_name)
        c = channel_indices[0]
        features_list.append([func(data[s, :, c],  fs=fs) for s in range(num_segments)])

        feature_labels.append(feature_name)
    features_array = np.column_stack(features_list)
    return features_array, np.array(feature_labels)





# -*- coding: utf-8 -*-

### ----------------------------- IMPORTS --------------------------- ###
import os
import click
from pathlib import Path
from seizyml.default_settings import (
    get_core_settings,
    get_user_settings,
    CustomOrderGroup,
    core_settings_path,
    user_settings_file,
    save_settings
)
### ----------------------------------------------------------------- ###

@click.group(cls=CustomOrderGroup)
@click.pass_context
def cli(ctx):
    """
    -------------------------------------------------------------
    
    \b                                                             
    \b                       _          ___  ___ _     
    \b                      (_)         |  \/  || |    
    \b              ___  ___  ______   _| .  . || |    
    \b             / __|/ _ \ |_  / | | | |\/| || |    
    \b             \__ \  __/ |/ /| |_| | |  | || |____
    \b             |___/\___|_/___|\__, \_|  |_/\_____/
    \b                             __/ |              
    \b                            |___/                                                  
    \b 

    --------------------------------------------------------------
                                                                                                                                           
    """
    ctx.obj={}
    core_settings = get_core_settings() 
    user_settings = get_user_settings(Path(core_settings['user_settings_path']))
    ctx.obj['core_settings'] = core_settings
    ctx.obj['user_settings'] = user_settings

# -------------------- MODEL HANDLING -------------------- #
@cli.command(name='train-model')
@click.option('--process', type=click.Choice(['compute_features', 'train_model'], case_sensitive=False), help='Choose to compute features, train model, or none for both.')
@click.pass_context
def train_models(ctx, process):
    """
    1: Train a new seizure detection model.
    """

    # Unpack settings
    usr, core = ctx.obj['user_settings'], ctx.obj['core_settings']

    # Check if model and model path exists
    if core['model_path'] and os.path.isfile(core['model_path']):
        overwrite = click.prompt(f"A model is already trained. Training will overwrite the model used. Proceed (y/n)?\n")
        if overwrite == 'n':
            click.secho("\n -> Operation cancelled. Training aborted.\n", fg='yellow')
            return
        elif overwrite != 'y':
            click.secho("\n -> Invalid input. Please enter 'y' for yes or 'n' for no.\n", fg='red')
            return
        
    # Get training path
    train_path = click.prompt("📂 Enter the training data path", type=str)
    if not os.path.exists(train_path):
        click.secho(f"❌ Train path '{train_path}' not found.", fg='red')
        return

    # Launch the GUI settings editor before training
    from seizyml.user_gui.settings_editor import edit_settings_gui
    click.echo("Opening settings editor GUI for updating settings...")
    updated_settings = edit_settings_gui(usr)
    usr = updated_settings
    click.echo("Settings updated via GUI.")
    
    # Train model and save settings
    from seizyml.train.train_models import train_model
    model_path = train_model(usr, train_path, process)
    core['model_path'] = model_path
    core['user_settings_path'] = os.path.join(train_path, user_settings_file)
    save_settings(core_settings_path, core)
    save_settings(core['user_settings_path'], usr)
    click.secho(f"✅ Created 'user settings' at {core['user_settings_path']}", fg='green')

@cli.command(name='select-model')
@click.argument('model_path')
@click.argument('user_settings_path')
@click.pass_context
def select_model(ctx, model_path, user_settings_path):
    """
    2: Select model trained model and user settings.
    """

    # Validate model file
    try:
        from joblib import load
        load(model_path)
        click.secho(f"✅ Model loaded successfully from '{model_path}'", fg='green')
    except Exception as e:
        click.secho(f"❌ Failed to load model: {e}", fg='red')
        return

    # Validate user settings file loading (The function checks internally)
    get_user_settings(Path(user_settings_path))
    click.secho(f"✅ Settings loaded successfully from '{user_settings_path}'", fg='green')

    # Update core settings
    core =  ctx.obj['core_settings']
    core['model_path'] = model_path
    core['user_settings_path'] = user_settings_path
    save_settings(core_settings_path, core)

@cli.command(name='set-datapath')
@click.argument('path')
@click.pass_context
def set_data_path(ctx, path):
    """
    3: Set the path to the dataset for analysis.
    """

    # Unpack settings
    core = ctx.obj['core_settings']
    usr = ctx.obj['user_settings']

    # Ensure a model is selected before setting the data path
    if not os.path.exists(core.get('model_path')):
        click.secho("❌ A model must be selected before setting the data path. Use 'select-model' first.", fg='red')
        return

    # Validate the provided path
    if not os.path.exists(path):
        click.secho(f"❌ The provided path '{path}' does not exist.", fg='red')
        return

    from seizyml.data_preparation.file_check import validate_data_structure
    # Data Validation
    click.secho(f"🔍 Validating data structure in '{path}'...", fg='cyan')
    data_validation_results = validate_data_structure(
        parent_path=path,
        data_dir=usr['data_dir'],
        model_channels=usr['channels'],
        model_fs=usr['fs'],
        model_win=usr['win']
    )
    if data_validation_results['errors']:
        for error in data_validation_results['errors']:
            click.secho(f"❌ {error}", fg='red')
        click.secho("❗ Data validation failed. Please fix the issues and try again.", fg='yellow')
        return
    else:
        click.secho("✅ Data structure validation successful!", fg='green')

    # Check processing status
    click.secho("📊 Checking data processing status...", fg='cyan')
    # run check for processed and model predictions
    from seizyml.data_preparation.file_check import check_processing_status
    checks = check_processing_status(path, 
                                    usr['data_dir'], 
                                    usr['processed_dir'], 
                                    usr['model_predictions_dir'])
    
    # update core settings
    core['parent_path'] = path
    core['data_validated'] = True
    core['is_processed'] = checks['is_processed']
    core['is_predicted'] = checks['is_predicted']
    save_settings(core_settings_path, core)
    
@cli.command(name='preprocess')
@click.pass_context
def preprocess(ctx):
    """
    4: Preprocess Data: Apply filtering and outlier removal.
    """
    # Unpack settings
    usr, core = ctx.obj['user_settings'], ctx.obj['core_settings']

    # Check if file validation has passed
    if not core.get('data_validated'):
        click.secho("❌ File validation has not passed. Please run `set-datapath` to validate data.", fg='red')
        return

    # Check if data has already been processed
    if core.get('is_processed', False):
        overwrite = click.prompt("⚠️ Files have already been processed. Do you want to overwrite? (y/n)", default='n')
        if overwrite.lower() != 'y':
            click.secho("🚫 Operation cancelled. Existing processed data will remain unchanged.", fg='yellow')
            return

    # Preprocessing paths
    load_path = os.path.join(core['parent_path'], usr['data_dir'])
    save_path = os.path.join(core['parent_path'], usr['processed_dir'])

    # Preprocess Data
    click.secho("⚙️ Preprocessing data: Filtering and removing large outliers...", fg='cyan')
    from seizyml.data_preparation.preprocess import PreProcess
    process_obj = PreProcess(load_path=load_path, save_path=save_path, fs=usr['fs'])
    process_obj.filter_data()

    # Update settings after successful processing
    core['is_processed'] = True
    save_settings(core_settings_path, core)
    click.secho("✅ Data preprocessing completed successfully!", fg='green')

@cli.command(name='predict')
@click.pass_context
def predict(ctx):
    """
    5: Generate Model Predictions.
    """

    # Unpack settings
    usr, core = ctx.obj['user_settings'], ctx.obj['core_settings']

    # Check if data has been preprocessed
    if not core.get('is_processed', False):
        click.secho("❌ Data needs to be preprocessed first. Please run `preprocess`.", fg='red')
        return

    # Check if a trained model is available
    if not core['model_path'] or not os.path.isfile(core['model_path']):
        click.secho(f"❌ No trained model found at {core['model_path']}. Please train or select a model first.", fg='red')
        return
    click.secho(f"📈 Using the following model for predictions: {core['model_path']}", fg='green')

    # Perform Predictions
    click.secho("🚀 Generating model predictions...", fg='cyan')
    from seizyml.data_preparation.get_predictions import ModelPredict
    model_obj = ModelPredict(
        model_path=core['model_path'],
        load_path=os.path.join(core['parent_path'], usr['processed_dir']),
        save_path=os.path.join(core['parent_path'], usr['model_predictions_dir']),
        channels=usr['channels'], win=usr['win'], fs=usr['fs'],
        post_processing_method=usr['post_processing_method'],
        dilation=usr['dilation'],
        erosion=usr['erosion'],
        event_threshold=usr['event_threshold'],
        boundary_threshold=usr['boundary_threshold'],
        rolling_window=usr['rolling_window']
    )
    model_obj.predict()

    # Update core settings after successful predictions
    core['is_predicted'] = True
    save_settings(core_settings_path, core)
    click.secho("✅ Model predictions generated successfully!", fg='green')

@cli.command(name='verify')
@click.pass_context
def verify(ctx):
    """
    6: Verify Detected Seizures.
    """

    # Unpack settings
    usr, core = ctx.obj['user_settings'], ctx.obj['core_settings']

    # Check if model predictions have been generated
    if not core.get('is_predicted', False):
        click.secho("❌ Model predictions have not been generated. Please run `predict` first.", fg='red')
        return

    # Validate that processed data and predictions exist
    click.secho("🔍 Validating predictions and processed data...", fg='cyan')
    import numpy as np
    from seizyml.data_preparation.file_check import check_verified
    verification_status = check_verified(
        folder=core['parent_path'],
        data_dir=usr['processed_dir'],
        csv_dir=usr['model_predictions_dir']
    )

    if verification_status:
        click.secho(f"❌ Verification failed. Missing data in: {verification_status}", fg='red')
        return

    # Launch User Verification GUI
    click.secho("🖥️ Launching seizure verification GUI...", fg='cyan')
    from seizyml.user_gui.user_verify import UserVerify
    verifier = UserVerify(
        parent_path=core['parent_path'],
        processed_dir=usr['processed_dir'],
        model_predictions_dir=usr['model_predictions_dir'],
        verified_predictions_dir=usr['verified_predictions_dir']
    )

    # User selects a file to verify
    file_id = verifier.select_file()

    # Retrieve data, seizure indices, and any existing verification colors
    data, idx_bounds = verifier.get_bounds(file_id, verified=False)
    color_array = None
    verified_path = os.path.join(verifier.verified_predictions_path, file_id)

    if os.path.exists(verified_path):
        try:
            color_array = np.loadtxt(os.path.join(verifier.verified_predictions_path, f'color_{file_id.replace(".csv", ".txt")}'), dtype=str)
        except Exception as e:
            click.secho(f"⚠️ Warning: Failed to load color array for {file_id}: {e}", fg='yellow')

    # Handle case with zero seizures
    if idx_bounds.shape[0] == 0:
        verifier.save_emptyidx(data.shape[0], file_id)
        click.secho(f"ℹ️ No seizures detected in {file_id}. Verification saved as empty.", fg='yellow')
    else:
        from seizyml.user_gui.verify_gui import VerifyGui
        VerifyGui(core['parent_path'], usr, file_id, data, idx_bounds, color_array)
        click.secho("✅ Verification process completed successfully!", fg='green')

@cli.command(name='extract-properties')
@click.pass_context
def extract_properties(ctx):
    """
    6: Extract Seizure Properties.
    """
    # Unpack settings
    usr, core = ctx.obj['user_settings'], ctx.obj['core_settings']

    # Check if verified predictions are available
    ver_path = os.path.join(core['parent_path'], usr['verified_predictions_dir'])
    if not os.path.exists(ver_path):
        click.secho("❌ Verified predictions not found. Please verify detected seizures first.", fg='red')
        return

    filelist = [f for f in os.listdir(ver_path) if f.endswith('.csv')]
    if not filelist:
        click.secho("❌ No verified seizures found. Please verify detected seizures.", fg='red')
        return

    # Extract properties
    click.secho("🔍 Extracting seizure properties...", fg='cyan')
    from seizyml.helper.get_seizure_properties import get_seizure_prop
    _, save_path = get_seizure_prop(core['parent_path'], usr['verified_predictions_dir'], usr['gui_win'])

    click.secho(f"✅ Seizure properties saved in '{save_path}'", fg='green')

@cli.command(name='feature-contribution')
@click.pass_context
def feature_contribution(ctx):
    """
    7: Plot Feature Contributions.
    """
    # Unpack settings
    core = ctx.obj['core_settings']

    # Check if model is trained
    if not core['model_path'] or not os.path.isfile(core['model_path']):
        click.secho(f"❌ Model not found at {core['model_path']}. Please train a model first.", fg='red')
        return

    # Load model and plot feature contributions
    click.secho("📊 Plotting feature contributions...", fg='cyan')
    from joblib import load
    import numpy as np
    import matplotlib.pyplot as plt

    model = load(core['model_path'])
    importances = np.abs(model.theta_[0] - model.theta_[1]) / (np.sqrt(model.var_[0]) + np.sqrt(model.var_[1]))
    importances = importances / np.sum(importances)

    plt.figure(figsize=(5, 3))
    ax = plt.gca()
    idx = np.argsort(importances)
    ax.barh(np.array(model.feature_labels)[idx], importances[idx], facecolor='#66bd7d', edgecolor='#757575')
    ax.spines['right'].set_visible(False)
    ax.spines['top'].set_visible(False)
    ax.set_xlabel('Feature Separation Score')
    ax.set_ylabel('Features')
    plt.tight_layout()
    plt.show()

if __name__ == '__main__':
    cli(obj={})



# -*- coding: utf-8 -*-         
               
### ------------------------ IMPORTS -------------------------------------- ###               
import os
import json
from tqdm import tqdm
import numpy as np
from sklearn.preprocessing import StandardScaler
from joblib import load
from seizyml.helper.io import load_data
from seizyml.helper.event_match import get_szr_idx, clean_predictions
from seizyml.helper.get_features import compute_selected_features
### ------------------------------------------------------------------------###


class ModelPredict:
    """
    Class for batch seizure prediction.

    """

    def __init__(self, model_path, load_path, save_path, channels, win, fs,
                   post_processing_method, dilation, erosion,
                  event_threshold, boundary_threshold, rolling_window):
        """
        Initializes the ModelPredict class.

        Parameters
        ----------
        model_path : str
            The path to the trained model file.
        load_path : str
            The path to the directory containing the input data.
        save_path : str
            The path to the directory where the output data will be saved.
        channels : list
            A list of channel names.
        win : int
            The window size in seconds.
        fs : float
            The sampling frequency in Hz.

        Returns
        -------
        None
        """
        # Set the input parameters as class attributes
        self.load_path = load_path
        self.save_path = save_path
        self.channels = channels
        self.win = win
        self.fs = fs
        
        # Load the trained model
        self.model = load(model_path)
        self.selected_features = self.model.feature_labels
        print('Model loaded:', self.model)
        
        # post-processing parameters
        self.operation = post_processing_method
        self.dilation = dilation
        self.erosion = erosion
        self.rolling_window = rolling_window
        self.t_high = event_threshold
        self.t_low = boundary_threshold

    def compute_metrics(self, file_id, y_pred, bounds_pred):
        """
        Computes and saves additional metrics such as the number of seizures detected 
        and the recording length for a given file to display during seizure verification.

        Parameters
        ----------
        file_id : str
            The file name with no extension.
        y_pred : 1D array
            The binary predictions.
        bounds_pred : 2D array
            The start and end points of detected seizures.

        Returns
        -------
        None
        """
        # Calculate the number of seizures detected
        num_seizures = bounds_pred.shape[0]
        
        # Calculate the recording length in hours
        recording_length = y_pred.shape[0] * self.win/60/60
        
        # Save these metrics in a dictionary
        metrics = {
            'file_id': file_id,
            'num_seizures': num_seizures,
            'recording_length': recording_length
        }

        # Write the metrics to a JSON file
        with open(os.path.join(self.save_path, file_id.replace('.h5','_metrics.json')), 'w') as f:
            json.dump(metrics, f)

    def predict(self):
        """
        Runs batch predictions.
        """
       
        print('---------------------------------------------------------------------------\n')
        print('---> Initiating Predictions.', '\n')
       
        # Create path prediction path
        if os.path.exists(self.save_path) is False:
            os.mkdir(self.save_path)
        
        # Get file list
        filelist = list(filter(lambda k: '.h5' in k, os.listdir(self.load_path)))
        
        # loop files (multilple channels per file)
        for i in tqdm(range(len(filelist)), desc = 'Progress'):
            
            # Get predictions (1D-array)
            y_pred, bounds_pred = self.get_feature_pred(filelist[i])

            # Compute and save additional metrics
            self.compute_metrics(filelist[i], y_pred, bounds_pred)
            
            # Convert prediction to binary vector and save as .csv
            ModelPredict.save_idx(os.path.join(self.save_path, filelist[i].replace('.h5','.csv')), y_pred, bounds_pred)
            
        print('---> Predictions have been generated for: ', self.save_path + '.','\n')
        print('---------------------------------------------------------------------------\n')
            
               
    def get_feature_pred(self, file_id):
        """
        Gets predictions for a given file.

        Parameters
        ----------
        file_id : str
            The file name with no extension.

        Returns
        -------
        y_pred : 1D array
            The binary predictions.
        bounds_pred : 2D array
            The start and end points of detected seizures.
        """
        
        # get data and true labels
        data = load_data(os.path.join(self.load_path, file_id))
        
        # Eextract features and normalize
        features, _ = compute_selected_features(data, self.selected_features, self.channels, self.fs)
        features = StandardScaler().fit_transform(features)
        
        # get clean predictions and event boundares
        y_pred = self.model.predict(features)
        clean_pred = clean_predictions(y_pred, operation=self.operation, 
                                       rolling_window=self.rolling_window,
                                       dilation=self.dilation, erosion=self.erosion,
                                       t_high=self.t_high, t_low=self.t_low)
        bounds_pred = get_szr_idx(clean_pred)
        
        return clean_pred, bounds_pred 

            
    def save_idx(file_path, y_pred, bounds_pred):
        """
        Saves user predictions to a CSV file as binary.

        Parameters
        ----------
        file_path : str
            The path to the file.
        y_pred : 1D array
            The binary predictions.
        bounds_pred : 2D array
            The start and end points of detected seizures.

        Returns
        -------
        None
        """
        # pre allocate file with zeros
        ver_pred = np.zeros(y_pred.shape[0])
        
        for i in range(bounds_pred.shape[0]):
            if bounds_pred[i,0] > 0:   
                ver_pred[bounds_pred[i,0]:bounds_pred[i,1]+1] = 1
                
        # save
        np.savetxt(file_path, ver_pred, delimiter=',', fmt='%i')

    
   



#### --------------- Imports --------------- ####
import click
from pathlib import Path
import yaml
#### --------------------------------------- ####

# -------------------- FILE PATHS -------------------- #
core_settings_file = 'core_settings.yaml'
user_settings_file = 'user_settings.yaml'
app_directory = Path(__file__).resolve().parent
core_settings_path = app_directory / core_settings_file

# -------------------- DEFAULT SETTINGS -------------------- #
def_core_settings = {
    'parent_path': '',
    'user_settings_path': '',
    'model_path': '',
    # app internal checks
    'data_validated': False,
    'is_processed': False,
    'is_predicted': False,
}

def_user_settings = {
    # data settings
    'channels': ['eeg1', 'eeg2'],
    'fs': 100,
    'win': 5,

    # gui settings
    'gui_win': 1,
    
    # features
    'features': [
        'line_length', 'kurtosis', 'skewness', 'rms', 'mad', 'var',
        'energy', 'hjorth_mobility', 'hjorth_complexity', 'delta_power',
        'theta_power', 'alpha_power', 'beta_power', 'gamma_power',
        'weighted_frequency', 'spectral_entropy', 'get_envelope_max_diff'
    ],
    'feature_select_thresh': 0.90,
    'feature_size': [5, 10, 15],
    'nleast_corr': 5,

    # post processing settings
    'post_processing_method': 'dual_threshold',
    'rolling_window': 6,
    'event_threshold': 0.5,
    'boundary_threshold': 0.2,
    'dilation': 2,
    'erosion': 1,

    # app directory names
    'data_dir': 'data',
    'processed_dir': 'processed',
    'model_predictions_dir': 'model_predictions',
    'verified_predictions_dir': 'verified_predictions',
    'trained_model_dir': 'models',
}

# -------------------- SETTINGS LOADER -------------------- #
def get_core_settings():
    """
    Load settings from a YAML file or initialize with defaults if the file doesn't exist.
    """
    try:
        if core_settings_path.is_file():
            with open(core_settings_path, "r") as file:
                settings = yaml.safe_load(file)
            return settings
        else:
            with open(core_settings_path, "w") as file:
                yaml.dump(def_core_settings, file)
            click.secho(f"✅ Created default 'core settings' at {core_settings_path}", fg='green')
            return def_core_settings
        
    except yaml.YAMLError as e:
        click.secho(f"❌ YAML error in {core_settings_path}: {e}", fg='red')
        return def_core_settings

def get_user_settings(settings_path):
    """
    Load settings from a YAML file or initialize with defaults if the file doesn't exist.
    """
    try:
        if settings_path.is_file():
            with open(settings_path, "r") as file:
                settings = yaml.safe_load(file)

            # Check for key consistency
            missing_keys = set(def_user_settings) - set(settings)
            if missing_keys:
                click.secho(f"⚠️  Missing keys in 'user_settings': {', '.join(missing_keys)}", fg='yellow')

            return settings
        else:
            click.secho(f"✅ Getting default 'user_settings'", fg='green')
            return def_user_settings

    except yaml.YAMLError as e:
        click.secho(f"❌ YAML error in {settings_path}: {e}", fg='red')
        return def_user_settings

# -------------------- SAVE SETTINGS FUNCTION -------------------- #
def save_settings(settings_path, settings):
    """
    Save settings to the specified YAML file.

    Parameters:
    - settings_path (str or Path): Path to the YAML file where settings will be saved.
    - settings (dict): The settings dictionary to be saved.

    Returns:
    - None
    """
    settings_path = Path(settings_path)
    with open(settings_path, 'w') as file:
        yaml.dump(settings, file)

# -------------------- CUSTOM CLICK GROUP -------------------- #
class CustomOrderGroup(click.Group):
    """
    A custom Click group that maintains the order of commands added to it.
    """
    def __init__(self, **attrs):
        super(CustomOrderGroup, self).__init__(**attrs)
        self.commands_in_order = []

    def command(self, *args, **kwargs):
        def decorator(f):
            cmd = super(CustomOrderGroup, self).command(*args, **kwargs)(f)
            self.commands_in_order.append(cmd.name)
            return cmd
        return decorator

    def list_commands(self, ctx):
        return self.commands_in_order



# -*- coding: utf-8 -*-

### -------------- IMPORTS -------------- ###
import numpy as np
from scipy.fftpack import fft
import scipy.stats as stats
### ------------------------------------- ###


# =============================================================================
#  Single Features
# =============================================================================

def get_psd(signal, fs):
    """
    Calculate Power spectral Density.

    Parameters
    ----------
    signal : 1D numpy array
    fs: sampling rate

    Returns
    -------
    freqs : 1D numpy array
    psd : 1D numpy array


    """
    # get power spectrum and normalize to signal length
    xdft = np.square(np.absolute(fft(signal)))
    xdft = xdft * (1/(fs*signal.shape[0]))
       
    # multiply *2 to conserve energy in positive frequencies
    psd = 2*xdft[0:int(xdft.shape[0]/2+1)]
    
    # generate corresponding frequency bins
    freqs = np.linspace(0, fs / 2, len(psd))
    return  freqs, psd

def get_power_area(signal, freq, fs):
    """
    Measure the power area of the signal.
    
    Parameters
    ----------
    signal : 1D numpy array
    freq: list(low, high)
    fs: sampling rate
    
    Returns
    -------
    power_area : np.float
    
    """
    
    freqs, psd = get_psd(signal, fs)
    flow = np.argmin(np.abs(freqs - freq[0]))
    fup = np.argmin(np.abs(freqs - freq[1]))
    return np.sum(psd[flow:fup+1])

def weighted_frequency(signal, fs):
    """
    Compute the weighted mean frequency (spectral centroid) of a signal.

    Parameters
    ----------
    signal : 1D numpy array
    fs: sampling rate
    
    Returns
    -------
    wm_frequency : np.float

    """
    freqs, psd = get_psd(signal, fs)
    wm_frequency = np.sum(freqs * psd) / np.sum(psd)
    return wm_frequency

def spectral_entropy(signal, fs):
    """
    Compute the spectral entropy of a normalized power spectral density (PSD).

    Parameters
    ----------
    psd : 1D numpy array
        Normalized power spectral density array.
    fs: sampling rate

    Returns
    -------
    spectral_ent : float
        Spectral entropy value.
    """
    _, psd = get_psd(signal, fs)
    psd_norm = psd / np.sum(psd)
    spectral_ent = stats.entropy(psd_norm)
    return spectral_ent

def delta_power(signal, fs):
    return get_power_area(signal, fs=fs, freq=[2, 4])

def theta_power(signal, fs):
    return get_power_area(signal, fs=fs, freq=[4.2, 8])

def alpha_power(signal, fs):
    return get_power_area(signal, fs=fs, freq=[8.2, 12])

def beta_power(signal, fs):
    return get_power_area(signal, fs=fs, freq=[12.2, 30])

def gamma_power(signal, fs):
    return get_power_area(signal, fs=fs, freq=[30.2, 50])


def line_length(signal, **kwargs):
    """
    Measures the line length of a signal.

    Parameters
    ----------
    signal : 1D numpy array
    -------
    line_length : np.float
    """
    return np.sum(np.abs(np.diff(signal))) / len(signal)

def skewness(signal, **kwargs):
    """
    Compute the skewness of a signal.

    Parameters
    ----------
    signal : 1D numpy array
        Input signal.

    Returns
    -------
    skewness_value : float
        Skewness of the signal.
    """
    return stats.skew(signal)

def kurtosis(signal, **kwargs):
    """
    Compute the kurtosis of a signal.

    Parameters
    ----------
    signal : 1D numpy array
        Input signal.

    Returns
    -------
    kurtosis_value : float
        Kurtosis of the signal.
    """
    return stats.kurtosis(signal)


def var(signal, **kwargs):
    """
    Measures variance of a signal.

    Parameters
    ----------
    signal : 1D numpy array
    """
    return np.var(signal)


def hjorth_mobility(signal, **kwargs):
    """
    Compute the Hjorth Mobility parameter of a signal.

    Parameters
    ----------
    signal : 1D numpy array
        Input signal.

    Returns
    -------
    mobility : float
        Hjorth Mobility of the signal.
    """
    if np.var(signal) == 0:
        return 0
    diff_signal = np.diff(signal)
    return np.sqrt(np.var(diff_signal) / np.var(signal))


def hjorth_complexity(signal, **kwargs):
    """
    Compute the Hjorth Complexity parameter of a signal.

    Parameters
    ----------
    signal : 1D numpy array
        Input signal.

    Returns
    -------
    complexity : float
        Hjorth Complexity of the signal.
    """
    if np.var(signal) == 0:
        return 0
    diff_signal = np.diff(signal)
    diff_diff_signal = np.diff(diff_signal)
    mobility = np.sqrt(np.var(diff_signal) / np.var(signal))
    complexity = np.sqrt(np.var(diff_diff_signal) / np.var(diff_signal)) / mobility
    return complexity

def rms(signal, **kwargs):
    """
    Measures root mean square of a signal.

    Parameters
    ----------
    signal : 1D numpy array
    """
    return np.sqrt(np.mean(np.square(signal)))

def max_envelope(signal, win, **kwargs):
    """
    Calculates max envelope of a signal across each window.

    Parameters
    ----------
    signal : 1D numpy array
    win : moving window
    Output
    -------
    max_envelope : 1D numpy array
    """
    env = np.zeros(signal.shape[0])
    for i in range(0,signal.shape[0]): 
        env[i] = np.max(signal[i:i+win])
    env[-win + 1:] = env[-win]
    return env

def get_envelope_max_diff(signal, **kwargs):
    """
    Measures the sum of the differences between the upper and lower envelopes.

    Parameters
    ----------
    signal : 1D numpy array
    win : moving window
    Output
    -------
    max_envelope : 1D numpy array
    """
    # fixed parameter
    win = 30 # moving window (0.3 seconds based on fs=100)
    up_env = max_envelope(signal, win)
    low_env = -max_envelope(-signal, win)  
    return np.sum(up_env - low_env)
    
def mad(signal, **kwargs):
    """
    Measures median absolute deviation of a signal.

    Parameters
    ----------
    signal : 1D numpy array
    """
    return np.median(np.abs(signal - np.mean(signal)))

def energy(signal, **kwargs):
    """
    Measures the energy of a signal.

    Parameters
    ----------
    signal : 1D numpy array
    -------
    energy : np.float
    """
    return np.sum(np.square(signal))




# -*- coding: utf-8 -*-

### ---------------- IMPORTS ------------------ ###
import os
import json
import tables
from pick import pick
import numpy as np
# User Defined
from seizyml.helper.event_match import get_szr_idx
### ------------------------------------------- ###


class UserVerify:
    """
        Class for user verification of detected seizures.
    """
    
    # class constructor (data retrieval)
    def __init__(self, parent_path, processed_dir, model_predictions_dir,
                 verified_predictions_dir):
        """
        
        Parameters
        ----------
        parent_path : str
        processed_dir : str
        model_predictions_dir : str
        verified_predictions_dir : str

        Returns
        -------
        None.

        """

        # set full paths 
        self.processed_path = os.path.join(parent_path, processed_dir)
        self.model_predictions_path = os.path.join(parent_path, model_predictions_dir)
        self.verified_predictions_path = os.path.join(parent_path, verified_predictions_dir)

        # make path if it doesn't exist
        if os.path.exists(self.verified_predictions_path) is False:
            os.mkdir(self.verified_predictions_path)

    def read_metrics(self, file_id):
        """
        Read metrics from json file.
        
        Parameters
        ----------
        file_id : str
            file id of file to read metrics from.
            
        Returns
        -------
        metrics : dict
            dictionary of metrics.
                
        """
        json_path = os.path.join(self.model_predictions_path, f"{file_id}_metrics.json")
        try:
            with open(json_path, 'r') as f:
                metrics = json.load(f)
                print(metrics)
        except FileNotFoundError:
            metrics = {'file_id': file_id, 'num_seizures': 'N/A', 'recording_length': 'N/A'}
        return metrics
        
    def select_file(self):
        """
        Select file to load from list. Adds stars next to files that have been scored already.
        
        Returns
        -------
        option : Str, selection of file id
        """
       
        # get all files in raw predictions folder 
        rawpredlist = list(filter(lambda k: '.csv' in k, os.listdir(self.model_predictions_path)))
       
        # get all files in user verified predictions
        verpredlist = list(filter(lambda k: '.csv' in k, os.listdir(self.verified_predictions_path)))
       
        # get unique list
        not_analyzed_filelist = list(set(rawpredlist) - set(verpredlist))
        
        # remaining filelist
        analyzed_filelist = list(set(rawpredlist) - set(not_analyzed_filelist))
        
        # filelist
        filelist = not_analyzed_filelist + analyzed_filelist

        # create display list
        display_list = []
        lists = [not_analyzed_filelist, analyzed_filelist]
        append_strs = ['','***']
        for append_str, filelists in zip(append_strs, lists):
            for file in filelists:
                metrics = self.read_metrics(file.replace('.csv',''))
                display_list.append(append_str + ' ' + str(metrics)[1:-1].replace("'",""))
        
        # select from command list
        title = 'Please select file for analysis: '
        display_list, index = pick(display_list, title, indicator = '-> ')

        return filelist[index]


    def get_bounds(self, file_id, verified):
        """
        Load data and calulate seizure bounds from predictions.

        Parameters
        ----------
        file_id : String

        Returns
        -------
        data : 3d Numpy Array (1D = segments, 2D = time, 3D = channel)
        idx_bounds : 2D Numpy Array (rows = seizures, cols = start and end points of detected seizures)
        verified: bool, True if file was verified

        """
        
        # Get predictions
        print('-> File being analyzed: ', file_id)
        if verified:
            pred_path = os.path.join(self.verified_predictions_path, file_id)
        else:
            pred_path = os.path.join(self.model_predictions_path, file_id)
        bin_pred = np.loadtxt(pred_path, delimiter=',', skiprows=0)
        idx_bounds = get_szr_idx(bin_pred)
        
        # load raw data for visualization
        data_path = os.path.join(self.processed_path, file_id.replace('.csv','.h5'))
        f = tables.open_file(data_path, mode='r')
        data = f.root.data[:]
        f.close()
        print('>>>>', idx_bounds.shape[0], 'seizures detected')
        
        return data, idx_bounds
            
    def save_emptyidx(self, data_len, file_id):
         """
         Save user predictions to csv file as binary
        
         Returns
         -------
         None.
        
         """
         # pre allocate file with zeros
         ver_pred = np.zeros(data_len)
         
         # save file
         np.savetxt(os.path.join(self.verified_predictions_path, file_id), ver_pred, delimiter=',',fmt='%i')
         print('Verified predictions for ', file_id, ' were saved\n')
    
    

       


#### ------------- Imports ------------- ####
import numpy as np
from scipy import ndimage
#### ----------------------------------- ####

def get_szr_idx(pred_array):
    """
    Identify seizure events and return their start and stop indices.
    
    Parameters
    ----------
    pred_array : numpy.ndarray, 1D boolean array where `True` indicates a seizure event.
        
    Returns
    -------
    idx_bounds : numpy.ndarray, 2D array of shape (n_events, 2) containing start and stop indices of valid events.
    
    Examples
    --------
    >>> pred_array = np.array([False, False, False, True, True, False, False])
    >>> find_szr_idx(pred_array)
    array([[3, 4]])
    """
    
    ref_pred = np.concatenate(([0], pred_array, [0]))
    transitions = np.diff(ref_pred)
    rising_edges = np.where(transitions == 1)[0]
    falling_edges = np.where(transitions == -1)[0] - 1
    idx_bounds = np.column_stack((rising_edges, falling_edges))

    return idx_bounds

def match_szrs_idx(bounds_true, y_pred):
    """
    Check for matching seizures in predictions based on ground-truth events.

    Parameters
    ----------
    bounds_true : numpy.ndarray, 2D array of start and stop indices for each true seizure event.
    y_pred : numpy.ndarray, 1D binary array with model's seizure predictions.

    Returns
    -------
    idx : numpy.ndarray
        1D binary array indicating matching seizure events.

    Example
    -------
    >>> bounds_true = np.array([[100, 150], [200, 250]])
    >>> y_pred = np.zeros(300)
    >>> y_pred[105:145] = 1
    >>> match_szrs_idx(bounds_true, y_pred)
    >>> array([ True, False])
    """
    idx = np.zeros(bounds_true.shape[0])
    for i in range(bounds_true.shape[0]):
        pred = y_pred[bounds_true[i, 0]:bounds_true[i, 1] + 1]
        if pred.sum() > 0:
            idx[i] = 1
    return idx.astype(bool)

### post-processing methods ###

# TODO add tests
def dual_threshold(raw_pred, t_high=.5, t_low=.2, win_size=6):
    """
    Apply hysteresis thresholding to detect events in a signal based on high and low thresholds.
    
    Hysteresis thresholding is a two-step process where a high threshold (`t_high`) is used to 
    initiate an event, and a lower threshold (`t_low`) is used to continue the event until 
    the signal falls below it. This method reduces noise and false positives in event detection.
    
    Parameters
    ----------
    raw_pred : numpy.ndarray
        1D array of raw predictions or signal values.
    t_high : float, optional, default=0.5
        High threshold value for starting an event.
    t_low : float, optional, default=0.2 
        Low threshold value for ending an event.
    win_size : int, optional, default=6
        Size of the rolling window (in samples) used to smooth the raw predictions 
        before applying thresholds.
    
    Returns
    -------
    numpy.ndarray
        1D binary array indicating detected events (1 for event, 0 otherwise).
    
    Notes
    -----
    - The rolling average is computed using a convolution operation to smooth the 
      raw predictions.
    - The binary propagation step ensures that once an event is triggered (based on 
      `t_high`), it continues until the signal drops below `t_low`.
    
    Examples
    --------
    >>> import numpy as np
    >>> from scipy import ndimage
    >>> raw_pred = np.array([0.1, 0.4, 0.6, 0.7, 0.5, 0.3, 0.1])
    >>> dual_threshold(raw_pred, t_high=0.5, t_low=0.2, win_size=3)
    array([0, 0, 1, 1, 1, 1, 0])
    
    Raises
    ------
    ValueError
        If `t_low` is greater than or equal to `t_high`.
    """
    
    mean_pred = np.convolve(raw_pred, np.ones(win_size)/win_size, mode='same')
    seeds = mean_pred >= t_high
    mask = mean_pred > t_low
    hysteresis_output = ndimage.binary_propagation(seeds, mask=mask)
    return hysteresis_output.astype(int)

def dilation_erosion(vector, dilation=0, erosion=0):
    """
    Applies morphological operations to a 1D binary vector.
    
    Parameters:
        vector (numpy.ndarray): Input 1D binary vector.
        dilation (int): Size of the dilation structuring element minus one (default is 0).
        erosion (int): Size of the erosion structuring element (default is 0).
    
    Returns:
        numpy.ndarray: The processed vector after applying the specified morphological operations, with padding removed.
        
    Notes
    -----
    Padding is added to the vector to prevent edge effects during morphological operations. 
    It is removed before returning the result.
    
    """
    padded_vector = np.concatenate(([0,0,0], vector, [0,0,0]))
    operation1 = ndimage.binary_closing(padded_vector, structure=np.ones(dilation+1)).astype(int)
    operation2 = ndimage.binary_opening(operation1, structure=np.ones(erosion+1)).astype(int)
    return operation2[3:-3]

def erosion_dilation(vector, dilation=0, erosion=0):
    """
    Applies morphological operations to a 1D binary vector.
    
    Parameters:
        vector (numpy.ndarray): Input 1D binary vector.
        dilation (int): Size of the dilation structuring element minus one (default is 0).
        erosion (int): Size of the erosion structuring element (default is 0).
    
    Returns:
        numpy.ndarray: The processed vector after applying the specified morphological operations, with padding removed.
        
    Notes
    -----
    Padding is added to the vector to prevent edge effects during morphological operations. 
    It is removed before returning the result.
    
    """
    padded_vector = np.concatenate(([0,0,0], vector, [0,0,0]))
    operation1 = ndimage.binary_opening(padded_vector, structure=np.ones(erosion+1)).astype(int)
    operation2 = ndimage.binary_closing(operation1, structure=np.ones(dilation+1)).astype(int)
    return operation2[3:-3]

def clean_predictions(raw_pred, operation='dual_threshold', dilation=None, erosion=None,
                      rolling_window=None, t_high=None, t_low=None):
    """
    Clean model predictions with post-processing methods.
    
    Parameters
    ----------
    raw_pred : numpy.ndarray
        1D binary array of raw model predictions.
    operation : str
        Type of post-processing to apply. Must be one of {'dilation_erosion', 'erosion_dilation', 'dual_threshold'}.
    dilation : int, optional
        Size of dilation structuring element. Required for `dilation_erosion` and `erosion_dilation`.
    erosion : int, optional
        Size of erosion structuring element. Required for `dilation_erosion` and `erosion_dilation`.
    rolling_window : int, optional
        Window size for smoothing in `dual_threshold`.
    t_high : float, optional
        High threshold for `dual_threshold`.
    t_low : float, optional
        Low threshold for `dual_threshold`.

    Returns
    -------
    numpy.ndarray
        1D binary array of cleaned predictions.

    Raises
    ------
    ValueError
        If an unsupported operation is provided or required parameters are missing.
    """
    if operation == 'dilation_erosion':
        clean_pred = dilation_erosion(raw_pred, dilation=dilation, erosion=erosion)
    elif operation == 'erosion_dilation':
        clean_pred = erosion_dilation(raw_pred, erosion=erosion, dilation=dilation)
    elif operation == 'dual_threshold':
        clean_pred = dual_threshold(raw_pred, win_size=rolling_window, t_high=t_high, t_low=t_low,)
    else:
        raise(f'Post-proccessing method {operation} was not found.')
    return clean_pred


if __name__ == '__main__':
    
    ### tests for clean predictions ###
    
    print('Testing dilation_erosion function...')
    # Test with provided example
    vector = np.array([1, 1, 0, 0, 1, 0, 1, 0, 0, 0, 0, 1, 1, 0, 0, 0, 0, 1, 0, 1, 0, 0, 1, 1, 0, 0, 1, 0, 0, 0, 1, 1, 0, 0, 0, 1])
    expect = np.array([1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 1, 1, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 1, 1, 0, 0, 0, 0])
    assert np.array_equal(dilation_erosion(vector, dilation=2, erosion=1), expect), "Test case 1 failed"

    # Test with all zeros
    vector = np.array([0, 0, 0, 0, 0, 0, 0, 0])
    expect = np.array([0, 0, 0, 0, 0, 0, 0, 0])
    assert np.array_equal(dilation_erosion(vector, dilation=2, erosion=1), expect), "Test case 2 failed"

    # Test with all ones
    vector = np.array([1, 1, 1, 1, 1, 1, 1, 1])
    expect = np.array([1, 1, 1, 1, 1, 1, 1, 1])
    assert np.array_equal(dilation_erosion(vector, dilation=2, erosion=1), expect), "Test case 3 failed"

    # Test with isolated ones
    vector = np.array([1, 0, 1, 0, 1, 0, 1, 0])
    expect = np.array([1, 1, 1, 1, 1, 1, 1, 0])
    assert np.array_equal(dilation_erosion(vector, dilation=2, erosion=1), expect), "Test case 4 failed"

    # Test with a single one
    vector = np.array([0, 0, 0, 1, 0, 0, 0])
    expect = np.array([0, 0, 0, 0, 0, 0, 0])
    assert np.array_equal(dilation_erosion(vector, dilation=2, erosion=1), expect), "Test case 5 failed"

    # Test with two ones close together
    vector = np.array([0, 1, 0, 1, 0])
    expect = np.array([0, 1, 1, 1, 0])
    assert np.array_equal(dilation_erosion(vector, dilation=2, erosion=1), expect), "Test case 6 failed"

    # Test with complex pattern
    vector = np.array([1, 0, 0, 1, 0, 0, 1, 0, 0, 1])
    expect = np.array([1, 1, 1, 1, 1, 1, 1, 1, 1, 1])
    assert np.array_equal(dilation_erosion(vector, dilation=2, erosion=1), expect), "Test case 7 failed"
    
    # Test with isolated ones
    vector = np.array([1, 0, 1, 0, 1, 0, 1, 1, 0])
    expect = np.array([1, 1, 1, 1, 1, 1, 1, 1, 0])
    assert np.array_equal(dilation_erosion(vector, dilation=2, erosion=1), expect), "Test case 8 failed"
    
    # Test with two blocks
    vector = np.array([1, 1, 0, 0, 0, 1, 0, 1, 0, 1, 1, 0])
    expect = np.array([1, 1, 0, 0, 0, 1, 1, 1, 1, 1, 1, 0])
    assert np.array_equal(dilation_erosion(vector, dilation=2, erosion=1), expect), "Test case 8 failed"
    print("All test cases passed!\n")
    
    print('Testing erosion_dilation function...')
    # Test with provided example
    vector = np.array([1, 1, 0, 0, 1, 0, 1, 0, 0, 0, 0, 1, 1, 0, 0, 0, 0, 1, 0, 1, 0, 0, 1, 1, 0, 1, 1, 0, 0, 0, 1, 1, 0, 0, 0, 1])
    expect = np.array([1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 0, 0, 0, 1, 1, 0, 0, 0, 0])
    assert np.array_equal(erosion_dilation(vector, dilation=2, erosion=1), expect), "Test case 1 failed"

    # Test with all zeros
    vector = np.array([0, 0, 0, 0, 0, 0, 0, 0])
    expect = np.array([0, 0, 0, 0, 0, 0, 0, 0])
    assert np.array_equal(erosion_dilation(vector, dilation=2, erosion=1), expect), "Test case 2 failed"

    # Test with all ones
    vector = np.array([1, 1, 1, 1, 1, 1, 1, 1])
    expect = np.array([1, 1, 1, 1, 1, 1, 1, 1])
    assert np.array_equal(erosion_dilation(vector, dilation=2, erosion=1), expect), "Test case 3 failed"

    # Test with isolated ones
    vector = np.array([1, 0, 1, 0, 1, 0, 1, 0])
    expect = np.array([0, 0, 0, 0, 0, 0, 0, 0])
    assert np.array_equal(erosion_dilation(vector, dilation=2, erosion=1), expect), "Test case 4 failed"

    # Test with a single pair
    vector = np.array([0, 0, 0, 1, 1, 0, 0])
    expect = np.array([0, 0, 0, 1, 1, 0, 0])
    assert np.array_equal(erosion_dilation(vector, dilation=2, erosion=1), expect), "Test case 5 failed"

    # Test with two ones close together
    vector = np.array([0, 1, 1, 0, 1, 1, 0])
    expect = np.array([0, 1, 1, 1, 1, 1, 0])
    assert np.array_equal(erosion_dilation(vector, dilation=2, erosion=1), expect), "Test case 6 failed"

    # Test with complex pattern
    vector = np.array([1, 1, 0, 1, 1, 0, 0, 1, 0, 1, 1])
    expect = np.array([1, 1, 1, 1, 1, 0, 0, 0, 0, 1, 1])
    assert np.array_equal(erosion_dilation(vector, dilation=2, erosion=1), expect), "Test case 7 failed"
    
    
    # Test with two blocks
    vector = np.array([1, 1, 0, 0, 0, 1, 1, 1, 0, 1, 1, 0])
    expect = np.array([1, 1, 0, 0, 0, 1, 1, 1, 1, 1, 1, 0])
    assert np.array_equal(erosion_dilation(vector, dilation=2, erosion=1), expect), "Test case 8 failed"
    print("All test cases passed!\n")
    
    ### Tests for get_szr_idx ###
    print('Testing get_szr_idx function...')
    # Test with provided example
    pred_array = np.array([0, 1, 0, 1, 1, 0, 1])
    expect = np.array([[1, 1], [3, 4], [6, 6]])
    assert np.array_equal(get_szr_idx(pred_array), expect), "Test case 1 failed"

    # Test with no seizures
    pred_array = np.array([0, 0, 0, 0])
    expect = np.array([]).reshape(0, 2)
    assert np.array_equal(get_szr_idx(pred_array), expect), "Test case 2 failed"

    # Test with continuous seizures
    pred_array = np.array([1, 1, 1, 1])
    expect = np.array([[0, 3]])
    assert np.array_equal(get_szr_idx(pred_array), expect), "Test case 3 failed"

    # Test with single seizure
    pred_array = np.array([0, 0, 1, 0, 0])
    expect = np.array([[2, 2]])
    assert np.array_equal(get_szr_idx(pred_array), expect), "Test case 4 failed"

    # Test with alternating seizures
    pred_array = np.array([0, 0, 1, 1, 1, 1, 0, 0, 0, 1, 1, 0, 0])
    expect = np.array([ [2, 5], [9, 10]])
    assert np.array_equal(get_szr_idx(pred_array), expect), "Test case 5 failed"

    # Test with multiple seizures
    pred_array = np.array([1, 1, 0, 0, 1, 1, 0, 1, 1, 0])
    expect = np.array([[0, 1], [4, 5], [7, 8]])
    assert np.array_equal(get_szr_idx(pred_array), expect), "Test case 6 failed"
    print("All test cases passed!\n")
    
    ### Tests for match_szrs_idx ###
    print('Testing match_szrs_idx function...')
   # Test with provided example
    bounds_true = np.array([[0, 3], [4, 6]])
    y_pred = np.array([1, 1, 0, 0, 1, 1, 1])
    expect = np.array([1, 1], dtype=bool)
    assert np.array_equal(match_szrs_idx(bounds_true, y_pred), expect), "Test case 1 failed"

    # Test with no matching seizures
    bounds_true = np.array([[0, 2], [4, 6]])
    y_pred = np.array([0, 0, 0, 0, 0, 0, 0])
    expect = np.array([0, 0], dtype=bool)
    assert np.array_equal(match_szrs_idx(bounds_true, y_pred), expect), "Test case 2 failed"

    # Test with all matching seizures
    bounds_true = np.array([[0, 1], [2, 3], [4, 5]])
    y_pred = np.array([1, 1, 1, 1, 1, 1])
    expect = np.array([1, 1, 1], dtype=bool)
    assert np.array_equal(match_szrs_idx(bounds_true, y_pred), expect), "Test case 3 failed"

    # Test with partially matching seizures
    bounds_true = np.array([[0, 2], [3, 5]])
    y_pred = np.array([0, 0, 0, 1, 0, 0])
    expect = np.array([0, 1], dtype=bool)
    assert np.array_equal(match_szrs_idx(bounds_true, y_pred), expect), "Test case 4 failed"

    # Test with empty bounds_true
    bounds_true = np.empty((0, 2), dtype=int)
    y_pred = np.array([1, 1, 1, 1])
    expect = np.array([], dtype=bool)
    assert np.array_equal(match_szrs_idx(bounds_true, y_pred), expect), "Test case 5 failed"

    # Test with y_pred all zeros
    bounds_true = np.array([[0, 2], [3, 5]])
    y_pred = np.array([0, 0, 0, 0, 0, 0])
    expect = np.array([0, 0], dtype=bool)
    assert np.array_equal(match_szrs_idx(bounds_true, y_pred), expect), "Test case 6 failed"
    
    # Test with some mseizures matching
    bounds_true = np.array([[0, 2], [3, 5]])
    y_pred = np.array([0, 0, 0, 0, 0, 0])
    expect = np.array([0, 0], dtype=bool)
    assert np.array_equal(match_szrs_idx(bounds_true, y_pred), expect), "Test case 7 failed"

    print("All test cases passed!")



####-------------------------------- Imports --------------------------- ######
import pytest
import numpy as np
from seizyml.helper.event_match import get_szr_idx, match_szrs_idx
####------------------------------- Fixtures --------------------------- ######

def create_bin_array(length, index):
    """
    Create binary array based on index.

    Parameters
    ----------
    length : int, array length
    index : 2D list, containing indx[[start1,stop1], [start2,stop2]]

    Returns
    -------
    arr : array, binary
    
    """
    arr = np.zeros(length)
    for i in index:
        arr[i[0]: i[1]+1] = 1
    return arr

@pytest.fixture
def properties():
    prop = {}
    return prop

@pytest.fixture
def create_binary_array():
    return create_bin_array


####--------------------------------------------------- ######


####---------------------------- Tests -------------------------- ######
@pytest.mark.parametrize("length, index", 
                          [(100, [[1, 5], [11, 15]]), 
                           (100, [[0, 4], [30, 60]]),
                           (100, [[1, 2], [15, 20], [90, 99]]),
                          ])
def test_get_szr_index(create_binary_array, length, index):
    
    # create array and find original bounds
    arr = create_binary_array(length, index)
    szr_bounds = get_szr_idx(arr)
    assert np.allclose(szr_bounds, np.array(index))

@pytest.mark.parametrize("length, index_true, index_pred, matching", 
                          [(100, [[1, 5], [11, 15], [50, 51]], [[4, 5], [12, 14]], 2), 
                           (100, [[0, 4], [30, 60]],  [[3, 4], [12, 15]], 1),
                           (100, [[0, 4], [30, 60]],  [[6, 7], [12, 15]], 0),
                           (100, [[1, 10], [15, 20], [90, 94]], [[1, 5], [18, 18], [91, 94]], 3),
                          ])
def test_get_szr_match_szrs_idx(create_binary_array, length,
                                 index_true, index_pred, matching):
    
    # get true bounds and pred array
    true_arr = create_binary_array(length, index_true)
    true_bounds = get_szr_idx(true_arr)
    pred_arr = create_binary_array(length, index_pred)
    
    # find matching index
    idx = match_szrs_idx(true_bounds, pred_arr)
    assert idx.sum() == matching


@pytest.mark.parametrize("length, index", 
                          [ (100, [[0, 99]]),  # Edge Case: Single segment
                            (100, [[0, 0], [99, 99]])  # Edge Case: 1's at edges
                          ])
def test_get_szr_index_edge_cases(create_binary_array, length, index):
    arr = create_binary_array(length, index)
    szr_bounds = get_szr_idx(arr)
    assert np.allclose(szr_bounds, np.array(index))








# -*- coding: utf-8 -*-

### -------------------------------- IMPORTS ------------------------------ ###
import os
import numpy as np
from tqdm import tqdm
from seizyml.helper.io import load_data, save_data
from scipy.signal import butter, zpk2sos, sosfiltfilt
### ------------------------------------------------------------------------###


###### --------------------- Batch process functions ----------------- ######

class PreProcess:
    """
    A class for preprocessing data by removing outliers and filtering data.

    Attributes
    ----------
    load_path : str
        The path to the directory containing the data to be preprocessed.
    save_path : str
        The path to the directory where the preprocessed data will be saved.
    fs : int
        The sampling frequency of the data.
    outier_threshold : int/float, optional
        The threshold for removing outliers. Default is 25.
    freq_cutoff : list, optional
        A list containing a single numeral for the high-pass filter cutoff frequency. Default is [2].

    Methods
    -------
    filter_data()
        Preprocesses the data in the parent folder.
    filter_clean(data)
        Filters and removes outliers from the data.

    """

    def __init__(self, load_path, save_path, fs, outier_threshold=25, freq_cutoff=[2]):
        """
        Initializes the PreProcess class.

        Parameters
        ----------
        load_path : str
            The path to the directory containing the data to be preprocessed.
        save_path : str
            The path to the directory where the preprocessed data will be saved.
        fs : int
            The sampling frequency of the data.
        outier_threshold : int/float, optional
            The threshold for removing outliers. Default is 25.
        freq_cutoff : list, optional
            A list containing a single numeral for the high-pass filter cutoff frequency. Default is [2].

        Returns
        -------
        None.

        """
        self.load_path = load_path
        self.save_path = save_path
        self.fs = fs
        self.outier_threshold = outier_threshold
        self.freq_cutoff = freq_cutoff

    def filter_data(self):
        """
        Preprocesses the data in the parent folder.

        Returns
        -------
        None.

        """
        # create save path if it doesn't exist
        if os.path.isdir(self.save_path) == 0:
            os.mkdir(self.save_path)

        # get file list
        filelist = list(filter(lambda k: '.h5' in k, os.listdir(self.load_path)))

        print('\n --->', len(filelist), 'files will be processed.\n')
        for i in tqdm(range(0, len(filelist)), desc='Progress:'): # loop through experiments

            # clean and filter data
            data = load_data(os.path.join(self.load_path, filelist[i]))
            data = self.filter_clean(data)

            # save clean data
            save_data(os.path.join(self.save_path, filelist[i]), data)

    def filter_clean(self, data):
        """
        Filters and removes outliers from the data.

        Parameters
        ----------
        data : 3d ndarray, (1d = segments, 2d = time, 3d = channels)
            The data to be filtered and cleaned.

        Returns
        -------
        data : 3d ndarray, (1d = segments, 2d = time, 3d = channels)
            The filtered and cleaned data.

        """
        dim = data.shape # get data dimensions

        if self.outier_threshold is not None: # if true remove outliers
            for i in range(data.shape[2]):
                temp = clean_signal(data[:,:,i].flatten(), threshold=self.outier_threshold)
                data[:,:,i] = temp.reshape((dim[0], dim[1]))

        if self.freq_cutoff is not None:
            for i in range(data.shape[2]):
                # high-pass filter data
                data[:,:,i] = batch_filter(data[:,:,i], butter_highpass, fs=self.fs, freq_cutoff=self.freq_cutoff)
        return data


def clean_signal(data, threshold=25):
    """
    Removes outliers and replaces with median

    Parameters
    ----------
    data : 1D signal, numpy array
    threshold : Real number, The default is 25.

    Returns
    -------
    clean_data : 1D signal, numpy array

    """
    
    clean_data = np.copy(data)
    clean_data = clean_data - np.mean(clean_data)
    idx = np.where(np.abs(data) > (threshold * np.std(data)))
    clean_data[idx] = np.median(clean_data)
    
    return clean_data

def batch_filter(data, filt_func, freq_cutoff, fs=100, verbose=False):
    """
    batch_filter(data)

    Parameters
    ----------
    data : 2d numpy array
    filt_func: filter object
    freq_cutoff: list with frequency cutoff(s)
    verbose: bool, if True verbose

    Returns
    -------
    filt_data : 2d numpy array

    """
    
    # Init data matrix
    filt_data = np.zeros(data.shape)
    filt_data = filt_data.astype(np.double)
    
    if verbose == True:
        for i in tqdm(range(data.shape[0])):
            filt_data[i,:] = filt_func(data[i,:], freq_cutoff, fs=fs)
            
    elif verbose == False:
        for i in range(data.shape[0]):
            filt_data[i,:] = filt_func(data[i,:], freq_cutoff, fs=fs)
        
        
    return filt_data

def butter_highpass(data, cutoff, fs, order=5):
    """
    butter_highpass(data, cutoff, fs, order = 5)

    Parameters
    ----------
    data : 1d ndarray, signal 
    cutoff : Float, cutoff frequency
    fs : Int, sampling rate
    order : Int, filter order. The default is 5.

    Returns
    -------
    y : 1d ndarray, filtered signal 

    """
    nyq = 0.5 * fs               # Nyquist Frequency (Hz)
    normal_cutoff = cutoff[0] / nyq # Low-bound Frequency (Normalised)
    
    # Design filter
    z,p,k = butter(order, normal_cutoff, btype='high', analog=False, output ='zpk') 
    
    sos = zpk2sos(z,p,k)         # Convert to second order sections
    y = sosfiltfilt(sos, data)   # Filter data
    return y








# -*- coding: utf-8 -*-

### ------------------------------- IMPORTS ----------------------------- ###
import os
import numpy as np
import pandas as pd
from uuid import uuid4
from tqdm import tqdm
import multiprocessing
from sklearn.model_selection import StratifiedKFold
from sklearn.model_selection import GridSearchCV
from joblib import dump
from seizyml.train.model_settings import models, hyper_params, metrics, fit_metric
njobs = int(multiprocessing.cpu_count()*.8)
### --------------------------------------------------------------------- ###

def get_feature_indices(selected_features, all_features):
    """
    Return the indices of selected features within the full set of features.
    
    Parameters
    ----------
    selected_features : list of str
        A list of feature names to locate within the complete feature set.
    all_features : numpy.ndarray
        A 1D array containing all available feature names.
        
    Returns
    -------
    numpy.ndarray
        An array of indices corresponding to the positions of each feature in 
        `selected_features` found in `all_features`.
        
    Raises
    ------
    IndexError
        If a feature in `selected_features` is not found in `all_features`.
    """

    # Loop through each selected feature to find its index in the all_features array
    indices = []
    for feature in selected_features:
        index = np.where(all_features == feature)[0][0]
        indices.append(index)

    return np.array(indices)

def grid_train(model_path, model_name, x, y, selected_features, feature_labels):
    """
    Perform hyperparameter tuning via grid search for a specified model on multiple feature sets,
    saving the best model based on F1 score per feature set.
    
    Parameters
    ----------
    model_path : str
        Directory path where the best model files will be saved.
    model_name : str
        The key name of the machine learning model in the `models` dictionary.
    x : numpy.ndarray
        A 2D array with shape (n_samples, n_features) containing the feature data.
    y : numpy.ndarray
        A 1D array containing the classification target labels.
    selected_features : dict
        A dictionary where each key is a feature set name and each value is a list of feature names
        (selected from `feature_labels`) to be used for that feature set.
    feature_labels : numpy.ndarray
        A 1D array of feature names corresponding to the columns in `x`.
    
    Returns
    -------
    pandas.DataFrame
        A DataFrame summarizing the grid search results with columns including:
        ['ID', 'HYPERPARAMETERS', ...], where additional columns represent performance metrics
        defined in the `metrics` dictionary.
    
    Notes
    -----
    - The best estimator for each feature set is saved as a joblib file using a unique ID.
    - The best index is computed using a custom transformation on the ranking of the refit metric.
    """

    # define Kfold 
    cv = StratifiedKFold(n_splits=5, random_state=2, shuffle=True)
    
    # iterate through feature sets
    data_list = []
    for feature_set in tqdm(selected_features, total=len(selected_features)):
        
        # perform grid search for particular feature set
        search = GridSearchCV(estimator=models[model_name],
                              param_grid=hyper_params[model_name],
                              n_jobs=njobs,
                              cv=cv,
                              verbose=0,
                              scoring=metrics,
                              refit=fit_metric)
        # get feature index
        sel_feature_idx = get_feature_indices(selected_features[feature_set], feature_labels)
        search.fit(x[:, sel_feature_idx], y)
        
        # create unique ID and save model with features labels
        uid = uuid4().hex
        best_model = search.best_estimator_
        best_model.feature_labels = selected_features[feature_set]
        dump(best_model, os.path.join(model_path, uid + '.joblib'))

        # get metrics and best parameters
        row = [uid, search.best_params_]
        best_metric_rank = 'rank_test_' + fit_metric
        results = search.cv_results_
        best_idx = np.argmin(results[best_metric_rank]-1)
        for m in metrics:
            scores = results['mean_test_' + m]
            row.append(scores[best_idx])
        data_list.append(row)
        
        df = pd.DataFrame(data=data_list, columns=['ID', 'HYPERPARAMETERS']+ list(metrics.keys()))
    return df

def train_and_save_models(trained_model_path, x, y, selected_features, feature_labels):
    """
    Train a collection of machine learning models using grid search over multiple feature sets,
    save the trained models, and return a summary DataFrame of the results.
    
    Parameters
    ----------
    trained_model_path : str
        Directory path where the trained model files will be saved. The directory will be created if it does not exist.
    x : numpy.ndarray or pandas.DataFrame
        A 2D array or DataFrame containing the feature data.
    y : numpy.ndarray or pandas.Series
        A 1D array or Series containing the target labels.
    selected_features : dict
        A dictionary mapping each feature set name to a list of selected feature names.
    feature_labels : numpy.ndarray
        A 1D array of feature names corresponding to the columns in `x`.
    
    Returns
    -------
    pandas.DataFrame
        A concatenated DataFrame summarizing the grid search results for all models,
        including unique model IDs, hyperparameters, and performance metrics, along with a 'model' column.
    
    Notes
    -----
    - This function iterates over all models defined in the `models` dictionary.
    """
    
    # create training dir
    if not os.path.exists(trained_model_path):
        os.makedirs(trained_model_path)
    
    # perform grid train, save best estimators
    df_list = []
    for m in models:
        print('Training model: ', m)
        train_df = grid_train(trained_model_path, m,  x, y, selected_features, feature_labels)
        train_df['model'] = m
        df_list.append(train_df)
   
    return pd.concat(df_list).reset_index(drop=True)

def train_model(usr, train_path, process):
    """
    Train a new seizure detection model using labeled training data.
    Depending on the `process` parameter, this function can compute features from raw data,
    train the model using grid search, or perform both steps sequentially.
    
    Parameters
    ----------
    usr : dict
        Dictionary of user settings that includes model parameters, feature extraction options,
        and data processing parameters.
    train_path : str
        Path to the training data directory containing raw data files ('.h5') and corresponding label files ('.csv').
    process : str or None
        Specifies the process to execute:
          - 'compute_features': Compute features from raw data.
          - 'train_model': Train the model using precomputed features.
          - Any other value or None: Execute both feature computation and model training.
    
    Returns
    -------
    str
        The file path to the saved best trained model (joblib file).
    
    Raises
    ------
    ValueError
        If the training data files are missing or if the file structure validation fails.
    
    Notes
    -----
    - When computing features, the raw data is preprocessed, features are computed and standardized,
      and then saved to disk.
    - Grid search is used for hyperparameter tuning, and the best model is selected based on the F1 score.
    """

    # imports
    import click
    from seizyml.train.train_models import train_and_save_models
    from seizyml.helper.io import load_data, save_data
    from seizyml.data_preparation.preprocess import PreProcess
    from sklearn.preprocessing import StandardScaler
    from seizyml.helper.get_features import compute_features
    from seizyml.train.select_features import select_features
    from tqdm import tqdm

    # Define process types
    process_type_options = ['compute_features', 'train_model']
    process_types = [process] if process in process_type_options else process_type_options

    if 'compute_features' in process_types:
        label_files = [f for f in os.listdir(train_path) if f.endswith('.csv')]
        h5_files = [f.replace('.csv', '.h5') for f in label_files]

        # Check data consistency
        from seizyml.data_preparation.file_check import train_file_check
        train_file_check(train_path, h5_files, label_files, usr['win'], usr['fs'], usr['channels'])

        # Compute features
        click.secho("⚙️ Computing features...", fg='green')
        x_all, y_all = [], []
        for h5_file, label_file in tqdm(zip(h5_files, label_files), total=len(h5_files)):
            x = load_data(os.path.join(train_path, h5_file))
            obj = PreProcess(None,None, fs=usr['fs'])
            x_clean = obj.filter_clean(x)
            features, feature_labels = compute_features(x_clean, usr['features'], usr['channels'], usr['fs'])
            features = StandardScaler().fit_transform(features)
            x_all.append(features)
            y_all.append(np.loadtxt(os.path.join(train_path, label_file)))

        # Save computed features
        save_data(os.path.join(train_path, 'features.h5'), np.concatenate(x_all))
        save_data(os.path.join(train_path, 'y.h5'), np.concatenate(y_all))
        np.savetxt(os.path.join(train_path, 'feature_labels.txt'), feature_labels, fmt='%s')

    if 'train_model' in process_types:
        click.secho("🎯 Training model...", fg='green')

        # load features, ground truth labels, and feature labels
        features = load_data(os.path.join(train_path, 'features.h5'))
        y = load_data(os.path.join(train_path, 'y.h5'))
        feature_labels = np.loadtxt(os.path.join(train_path, 'feature_labels.txt'), dtype=str)

        # select feature sets
        selected_features = select_features(features, y, feature_labels,
                                            r_threshold=usr['feature_select_thresh'],
                                            feature_size=usr['feature_size'],
                                            nleast_correlated=usr['nleast_corr'])

        # train models and find the one with best f1 score
        trained_model_path = os.path.join(train_path, usr['trained_model_dir'])
        train_df = train_and_save_models(trained_model_path, features, y, selected_features, feature_labels)
        train_df.to_csv(os.path.join(trained_model_path, 'trained_models.csv'), index=False)
        best_model_id = train_df.loc[train_df['F1'].idxmax(), 'ID']
        model_path = os.path.join(trained_model_path, f'{best_model_id}.joblib')

        return model_path




# -*- coding: utf-8 -*-
##### ----------------------------- IMPORTS ----------------------------- #####
import numpy as np
import pandas as pd
from sklearn.feature_selection import f_classif
##### ------------------------------------------------------------------- #####


def select_features(x, y_true, feature_labels, r_threshold, feature_size, nleast_correlated=0):
    """
     Selects features based on ANOVA F-values, while dropping highly correlated features 
     according to their ANOVA ranks, and optionally includes the least correlated features.
    
     Parameters:
     - x (array-like): The feature matrix with shape (n_samples, n_features).
     - y_true (array-like): The true class labels with shape (n_samples,).
     - feature_labels (array-like): The feature labels corresponding to columns from x.
     - r_threshold (float): Correlation threshold for feature selection. Features with correlation 
       higher than this threshold will be considered for dropping.
     - feature_size (list of int): List of sizes for the feature subsets to be selected.
     - nleast_correlated (int): Number of least correlated features to be included 
       in the final selection.
    
     Returns:
     - feature_dict (dict): A dictionary with keys corresponding to the different feature subsets.
       Each key has two variations:
         - 'best_{size}': Top features based solely on ANOVA ranks.
         - 'best_{size}_and_leastcorr': Top features combined with the least correlated features.
    """
    
    # Calculate correlation matrix
    corr_matrix = np.corrcoef(x.T)
    corr_df = pd.DataFrame(corr_matrix, index=feature_labels, columns=feature_labels)

    # Perform ANOVA
    f_vals_anova, _ = f_classif(x, y_true)
    anova_ranks = pd.Series(f_vals_anova, index=feature_labels).rank(method='min', ascending=False)
    
    # Identify highly correlated feature pairs
    upper_triangle = np.triu(np.ones(corr_df.shape), k=1).astype(bool)
    high_corr_pairs = corr_df.where(upper_triangle).stack()
    high_corr_pairs = high_corr_pairs[high_corr_pairs > r_threshold].index.tolist()

    # Determine which feature to drop based on lower ANOVA rank
    to_drop = set()
    for feature1, feature2 in high_corr_pairs:
        if anova_ranks[feature1] > anova_ranks[feature2]:
            to_drop.add(feature1)
        else:
            to_drop.add(feature2)
    selected_features = [str(f) for f in feature_labels if f not in to_drop]
    
    # Use the original ANOVA ranks to sort the selected features
    selected_anova_ranks = anova_ranks[selected_features]

    # Find the n least correlated features to the top 5 features
    top_5_features = selected_anova_ranks.nsmallest(5).index.values
    corr_to_top_5 = corr_df.loc[:, top_5_features].abs().mean(axis=1)
    least_corr_features = corr_to_top_5.nsmallest(nleast_correlated).index.tolist()

    # Create the feature dictionary for the specified feature sizes
    feature_dict = {}
    for size in feature_size:
        top_features = selected_anova_ranks.nsmallest(size).index.tolist()
        feature_dict[f'best_{str(size)}'] = top_features
        
    if nleast_correlated > 0:
        for size in feature_size:
            top_features = selected_anova_ranks.nsmallest(size).index.tolist()
            feature_dict[f'best_{str(size)}_and_leastcorr'] =  top_features + [f for f in least_corr_features if f not in top_features]

    return feature_dict


    
    





# -*- coding: utf-8 -*-

### -------------------------------- IMPORTS ------------------------------ ###
import os
from tqdm import tqdm
from seizyml.helper.io import load_data
### ------------------------------------------------------------------------###

def train_file_check(train_path, h5_files, label_files, win, fs, channels):
    """
    Check the existence and structure of H5 and label CSV files.

    Parameters
    ----------
    train_path : str
        Path to the directory containing training files.
    h5_files : list
        List of H5 file names.
    label_files : list
        List of corresponding CSV label file names.
    win : int or float
        The window size in seconds.
    fs : int or float
        The sampling frequency in Hz of the data.
    channels : int, optional
        The expected number of channels in the H5 files. The default is 2.

    Raises
    ------
    ValueError
        If any file is missing or improperly structured.
    """
    for x_path, y_path in zip(h5_files, label_files):
        x_full_path = os.path.join(train_path, x_path)
        y_full_path = os.path.join(train_path, y_path)

        if not os.path.exists(x_full_path) or not os.path.exists(y_full_path):
            raise ValueError(f'❌ Missing file: {x_path if not os.path.exists(x_full_path) else y_path}')

        x = load_data(x_full_path)
        if x.shape[2] != len(channels):
            raise ValueError(f'❌ Channel mismatch: Expected {len(channels)} channels, found {x.shape[2]} in {x_path}')

        if x.shape[1] != int(fs * win):
            raise ValueError(f'❌ Window size mismatch: Expected {int(fs * win)} samples, found {x.shape[1]} in {x_path}')

def validate_data_structure(parent_path, data_dir, model_channels, model_fs, model_win):
    """
    Validate the data directory structure and ensure it matches the model configuration.
    """
    errors = []
    data_path = os.path.join(parent_path, data_dir)

    # Check if data directory exists
    if not os.path.exists(data_path):
        errors.append(f"Data directory '{data_dir}' not found in '{parent_path}'.")
        return {'errors': errors}

    # Check H5 files
    from seizyml.helper.io import load_data
    h5_files = [f for f in os.listdir(data_path) if f.endswith('.h5')]

    if not h5_files:
        errors.append("No H5 files found in the data directory.")
    else:
        for h5_file in h5_files:
            try:
                data = load_data(os.path.join(data_path, h5_file))
                if data.shape[2] != len(model_channels):
                    errors.append(f"❌ {h5_file}: Channel mismatch (Expected {len(model_channels)}, Found {data.shape[2]}).")
                if data.shape[1] != model_fs * model_win:
                    errors.append(f"❌ {h5_file}: Window size mismatch (Expected {model_fs * model_win}, Found {data.shape[1]}).")
            except:
                errors.append(f"❌ {h5_file}: Failed to load H5 file.")

    return {'errors': errors}

def check_processing_status(parent_path, data_dir, processed_dir, model_predictions_dir):
    """
    Check the integrity of the main directories and their contents.

    Parameters
    ----------
    parent_path : str
        The parent directory containing the data, processed, and model predictions directories.
    data_dir : str
        The name of the data directory containing H5 data files.
    processed_dir : str
        The name of the processed directory.
    model_predictions_dir : str
        The name of the model predictions directory.

    Returns
    -------
    dict:
        processed_check : bool
            True if the processed directory exists and has the same files as the data directory, False otherwise.
        model_predictions_check : bool
            True if the model predictions directory exists and has the same files as the processed directory, False otherwise.

    """
    
    # initiate check variables for processed and model predictions 
    processed_check = True
    model_predictions_check = True
    
    # get paths for h5 data, processed and model predictions
    h5_path = os.path.join(parent_path, data_dir)
    processed_path = os.path.join(parent_path, processed_dir)
    model_predictions_path = os.path.join(parent_path, model_predictions_dir)

    # check if paths exist    
    if not os.path.exists(processed_path):
        processed_check = False
    if not os.path.exists(model_predictions_path):
        model_predictions_check = False

    if processed_check:
        h5 = set(x.replace('.h5', '') for x in os.listdir(h5_path))
        processed = set(x.replace('.h5', '') for x in os.listdir(processed_path))
        if h5 != processed:
            processed_check = False
            
    if processed_check and model_predictions_check:
        model_predictions = {x.replace('.csv', '') for x in os.listdir(model_predictions_path) if x[-4:] == '.csv'}   
        if processed != model_predictions:
            model_predictions_check = False
        
    return {'is_processed':processed_check, 'is_predicted':model_predictions_check}

def check_verified(folder, data_dir, csv_dir):
    """
    Check if folders exist and if h5 files match csv files.

    Parameters
    ----------
    folder : dict, with config settings
    data_dir : str, data directory name
    true_dir : str, csv directory name

    Returns
    -------
    None,str, None if test passes, otherwise a string is returned with the name
    of the folder where the test did not pass

    """
    
    h5_path = os.path.join(folder, data_dir)
    ver_path = os.path.join(folder, csv_dir)
    if not os.path.exists(h5_path):
        return h5_path
    if not os.path.exists(ver_path):
        return ver_path
    h5 = {x.replace('.h5', '') for x in os.listdir(h5_path)}
    ver = {x.replace('.csv', '') for x in os.listdir(ver_path)}   
    if len(h5) != len(h5 & ver):
        return folder 




# -*- coding: utf-8 -*-

### ------------------------------- IMPORTS ----------------------------- ###
import numpy as np
from sklearn.naive_bayes import GaussianNB
### --------------------------------------------------------------------- ###


# define metrics and Kfold partitions
fit_metric = 'BALANCED_ACCURACY'
metrics = {'AUC':'roc_auc', 'BALANCED_ACCURACY':'balanced_accuracy', 'PRECISION':'precision', 'RECALL':'recall', 'F1':'f1'}

# define models and hyperparameters
models = {  
            'gaussian_nb' : GaussianNB(),          
          }

hyper_params = {
                
                'gaussian_nb': {
                                'var_smoothing': np.logspace(-2,-8, num=7)},

                }